import warnings
import numpy as np
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf

from itertools import combinations
from sklearn.base import BaseEstimator


class EmpiricalMeanSL(BaseEstimator):
    """Empirical mean estimator in the format of SciKit learn. This estimator is for use with the SuperLearner
    functionality.

    Note
    ----
    Generally, I do not recommend its use outside of SuperLearner. Essentially the empirical mean is a
    baseline estimator with which to compare other estimators included in SuperLearner.

    Parameters
    ----------
    None

    Examples
    --------
    Setup the environment and data set

    >>> from zepid import load_sample_data
    >>> from zepid.superlearner import EmpiricalMeanSL
    >>> df = load_sample_data(False).dropna()
    >>> X = np.asarray(df[['art', 'male', 'age0']])
    >>> y = np.asarray(df['dead'])

    EmpiricalMean estimation

    >>> emp_mean = EmpiricalMeanSL()
    >>> emp_mean.fit(X=X, y=y)

    EmpiricalMean prediction

    >>> emp_mean.predict(X=X)
    """

    def __init__(self):
        self.empirical_mean = np.nan

    def fit(self, X, y):
        """Estimate the empirical mean based on X and y. While X in input, it has no effect on the estimated empirical
        mean (since the empirical mean is the average for the full y).

        Parameters
        ----------
        X : numpy.array
            Training data
        y : numpy.array
            Target values

        Returns
        -------
        None
        """
        # Error Checking
        if X.shape[0] != y.shape[0]:
            raise ValueError("X and y must have the same number of observations (rows).")
        if np.any(np.isnan(X)) or np.any(np.isnan(y)):
            raise ValueError("It looks like there is missing values in X or y. EmpiricalMeanSL does not support "
                             "missing data.")

        self.empirical_mean = np.mean(y)
        return self

    def predict(self, X):
        """Predict the value of y given a set of X covariates. Because X has no effect on the empirical mean, the
        mean from the data used in the fit() step is returned for all observations.

        Parameters
        ----------
        X : numpy.array
            NumPy array of covariates

        Returns
        -------
        NumPy array of predicted empirical means of the dimension X.shape[0]
        """
        return np.array([self.empirical_mean] * X.shape[0])


class GLMSL:
    """Generalized Linear Model for use with SuperLearner. This function is a wrapper function for the statsmodels
    `GLM` class. This is because the GLM implementation in statsmodels is not natively compatible with the sklearn /
    SuperLearner class. Compatible with all options available in the statsmodels families.

    Parameters
    ----------
    family: statsmodels.families.family
        Family to use for the model. All statsmodels supported families are also supported
    verbose : bool, optional

    Examples
    --------
    Setup the environment and data set

    >>> import statsmodels.api as sm
    >>> from zepid import load_sample_data
    >>> from zepid.superlearner import GLMSL
    >>> df = load_sample_data(False).dropna()
    >>> X = np.asarray(df[['art', 'male', 'age0']])
    >>> y = np.asarray(df['dead'])

    GLMSL example (logit model)

    >>> f = sm.families.family.Binomial()
    >>> glm = GLMSL(family=f)
    >>> glm.fit(X, y)
    """
    def __init__(self, family, verbose=False):
        self._family_ = family
        self._verbose_ = verbose

        # Storage items
        self.model = None

    def fit(self, X, y):
        """Estimate the GLM

        Parameters
        ----------
        X : numpy.array
            Training data
        y : numpy.array
            Target values

        Returns
        -------
        None
        """
        # Error Checking
        if X.shape[0] != y.shape[0]:
            raise ValueError("X and y must have the same number of observations (rows).")
        if np.any(np.isnan(X)) or np.any(np.isnan(y)):
            raise ValueError("It looks like there is missing values in X or y. GLMSL does not support "
                             "missing data.")

        # Final results
        self.model = sm.GLM(y, np.hstack([np.zeros([X.shape[0], 1]) + 1, X]),  # Adds intercept into model
                            family=self._family_).fit()
        if self._verbose_:
            print(self.model.summary())

        return self

    def predict(self, X):
        """Predict using the fitted GLM.

        Parameters
        ----------
        X : numpy.array
            Samples following the same pattern as the X array input into the fit() statement.

        Returns
        -------
        Returns predicted values from the GLM
        """
        # Adding intercept
        Xd = np.hstack([np.zeros([X.shape[0], 1]) + 1, X])

        # Generating predictions to return
        return self.model.predict(Xd)

    def get_params(self, deep=True):
        """For sklearn.base.clone() compatibility"""
        return {"family": self._family_,
                "verbose": self._verbose_}

    def set_params(self, **parameters):
        """For sklearn.base.clone() compatibility"""
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self


class StepwiseSL:
    """Step-wise model selection for Generalized Linear Model selection for use with SuperLearner. Briefly, each
    combination of models is compared by AIC with the best one selected. The model selection procedure continues until
    there are no improvements in the model by AIC. The optimal is the best model estimated by the step-wise selection
    procedure and the lowest AIC value.

    Parameters
    ----------
    family: statsmodels.families.family
        Family to use for the model. All statsmodels supported families are also supported
    selection : str, optional
        Method of step-wise selection to use. Options are `'forward'` and `'backward'`. Default is backward, which
        starts from the full model inclusion and removes terms one at a time.
    order_interaction : int, optional
        Order of interactions to explore. For example, `interaction_order=0` explores only the main effects.
    verbose : bool, optional

    Examples
    --------
    Setup the environment and data set

    >>> import statsmodels.api as sm
    >>> from zepid import load_sample_data
    >>> from zepid.superlearner import StepwiseSL
    >>> df = load_sample_data(False).dropna()
    >>> X = np.asarray(df[['art', 'male', 'age0']])
    >>> y = np.asarray(df['dead'])

    StepwiseSL estimation with no interactions

    >>> f = sm.families.family.Binomial()
    >>> step_sl = StepwiseSL(family=f, method="backward", order_interaction=0)
    >>> step_sl.fit(X, y)

    StepwiseSL prediction

    >>> step_sl.predict(X=X)

    StepwiseSL with all first-order interactions

    >>> step_sl = StepwiseSL(family=f, method="backward", order_interaction=1)
    >>> step_sl.fit(X, y)

    StepwiseSL with forward selection and all second-order interactions

    >>> step_sl = StepwiseSL(family=f, method="forward", order_interaction=2)
    >>> step_sl.fit(X, y)
    """
    def __init__(self, family, selection="backward", order_interaction=0, verbose=False):
        # Error Checking
        if selection.lower() not in ["forward", "backward"]:
            raise ValueError("`method` must be one of the following: backward, forward")
        if order_interaction < 0 or type(order_interaction) is not int:
            raise ValueError("interaction_order must be a non-negative integer")

        self._family_ = family
        self._verbose_ = verbose
        self._selection_ = selection
        self._order_ = order_interaction

        # Storage items
        self.model_optim = None
        self.cols_optim = None

    def fit(self, X, y):
        """Estimate the optimal GLM

        Parameters
        ----------
        X : numpy.array
            Training data
        y : numpy.array
            Target values

        Returns
        -------
        None
        """
        # Error Checking
        if X.shape[0] != y.shape[0]:
            raise ValueError("X and y must have the same number of observations (rows).")
        if np.any(np.isnan(X)) or np.any(np.isnan(y)):
            raise ValueError("It looks like there is missing values in X or y. StepWiseSL does not support "
                             "missing data.")
        if X.shape[1] < self._order_:  # Checking X shape with order_interaction
            warnings.warn("order_interaction is greater than the number of columns. This is not possible to assess, "
                          "so order_interaction="+str(int(X.shape[1]))+" will be assessed instead.", UserWarning)
            self._order_ = X.shape[1]

        # Creating all x-order interaction terms for assessment
        Xu = self._all_order_interactions_(X, self._order_)

        # Determining method of selection
        if self._selection_ == "backward":
            # Estimating full model as starting point
            full_model = sm.GLM(y, np.hstack([np.zeros([X.shape[0], 1]) + 1, Xu]),  # Adds intercept into model
                                family=self._family_).fit()
            if np.isnan(full_model.aic):
                raise ValueError("Saturated model is having trouble converging. Reduce the number of covariates, the "
                                 "order_interaction, or use selection=forward instead")
            best_aic = full_model.aic
            if self._verbose_:
                print(full_model.summary())
                print("Full-Model AIC:", best_aic)

            # Determining best AIC via backwards step-wise selection
            best_cols = list(range(Xu.shape[1]))
            best_alt_aic = best_aic
            best_alt_model = full_model
            best_alt_cols = best_cols.copy()

            while best_aic >= best_alt_aic and len(best_cols) - 1 >= -1:
                best_aic = best_alt_aic
                best_model = best_alt_model
                best_cols = best_alt_cols
                if self._verbose_:
                    print("\nCurrent Optim:", best_cols)

                if len(best_cols) - 1 == -1:  # necessary break for intercept-only to be outpute correctly
                    break
                if self._verbose_:
                    print("\nValid Combinations...")
                alt_models = list(combinations(best_cols, len(best_cols) - 1))
                best_alt_model, best_alt_cols, best_alt_aic = None, None, np.inf
                for alt in alt_models:
                    alt_model = sm.GLM(y, np.hstack([np.zeros([X.shape[0], 1]) + 1, Xu[:, alt]]),  # Adds intercept
                                       family=self._family_).fit()
                    if self._verbose_:
                        print("Columns:", alt)
                        print("AIC:    ", alt_model.aic)
                    if alt_model.aic < best_alt_aic:
                        best_alt_model = alt_model
                        best_alt_aic = alt_model.aic
                        best_alt_cols = alt

        # Determining method of selection
        if self._selection_ == "forward":
            # Estimating null model as starting point
            null_model = sm.GLM(y, np.zeros([X.shape[0], 1]) + 1,  # intercept-only model
                                family=self._family_).fit()
            best_aic = null_model.aic
            if self._verbose_:
                print(null_model.summary())
                print("Full-Model AIC:", best_aic)

            # Determining best AIC via forwards step-wise selection
            best_cols = ()
            best_alt_aic = best_aic
            best_alt_model = null_model
            best_alt_cols = ()
            vars_to_select = list(range(Xu.shape[1]))
            best_alt_var = None

            while best_aic >= best_alt_aic and len(best_cols) <= Xu.shape[1]:
                best_aic = best_alt_aic
                best_model = best_alt_model
                best_cols = best_alt_cols

                # Removing the previous best variable from those to add
                vars_to_select = list(vars_to_select)
                try:
                    vars_to_select.remove(best_alt_var)
                except ValueError:
                    pass
                vars_to_select = tuple(vars_to_select)

                if self._verbose_:
                    print("\nCurrent Optim:", best_cols)

                if len(best_cols) == Xu.shape[1]:  # necessary break for saturated to be output correctly
                    break
                if self._verbose_:
                    print("\nValid Combinations...")
                best_alt_model, best_alt_cols, best_alt_aic = None, None, np.inf
                for var in vars_to_select:
                    alt = best_cols + (var, )
                    alt_model = sm.GLM(y, np.hstack([np.zeros([X.shape[0], 1]) + 1, Xu[:, alt]]),
                                       family=self._family_).fit()
                    if self._verbose_:
                        print("Columns:", alt)
                        print("AIC:    ", alt_model.aic)
                    if alt_model.aic < best_alt_aic and not np.isnan(alt_model.aic):
                        best_alt_model = alt_model
                        best_alt_aic = alt_model.aic
                        best_alt_cols = alt
                        best_alt_var = var

        # Final results
        self.model_optim = best_model
        self.cols_optim = best_cols
        if self._verbose_:
            print(self.model_optim.summary())

        return self

    def predict(self, X):
        """Predict using the optimal GLM, where optimal is defined as the lowest AIC for the step-wise selection
        procedure used.

        Parameters
        ----------
        X : numpy.array
            Samples following the same pattern as the X array input into the fit() statement. All order_interaction
            terms are created in this step for the input X (i.e. the user does not need to create any of the x-order
            interaction terms)

        Returns
        -------
        Returns predicted values from the optimal GLM
        """
        # Creating all x-order interaction terms for assessment
        Xu = self._all_order_interactions_(X, self._order_)

        # Adding intercept
        Xd = np.hstack([np.zeros([X.shape[0], 1]) + 1, Xu[:, self.cols_optim]])

        # Generating predictions to return
        return self.model_optim.predict(Xd)

    @staticmethod
    def _all_order_interactions_(X, x_order):
        """Background function

        Parameters
        ----------
        X : numpy.array
            Input data array
        x_order : int
            Order of interactions to generate. x_order=1 generates all first order interactions

        Returns
        -------
        Array containing the original X and all corresponding x-order interactions
        """
        Xu = X.copy()
        assessed_order = 1
        while x_order >= assessed_order:
            assessed_order += 1

            # Generating unique combinations of columns
            combos_order = list(combinations(range(X.shape[1]), assessed_order))

            # Creating interaction term for all unique combinations
            for co in combos_order:
                interaction_term = np.prod(X[:, co], axis=1).reshape((X.shape[0], 1))
                Xu = np.hstack([Xu, interaction_term])

        return Xu

    def get_params(self, deep=True):
        """For sklearn.base.clone() compatibility"""
        return {"family": self._family_,
                "selection": self._selection_,
                "order_interaction": self._order_,
                "verbose": self._verbose_}

    def set_params(self, **parameters):
        """For sklearn.base.clone() compatibility"""
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self



import warnings
from collections import namedtuple

import numpy as np
from scipy.stats import norm

Results = namedtuple('Results',
                     ['point_estimate', 'lower_bound', 'upper_bound', 'standard_error', 'alpha', 'measure']
                     )


def normal_ppf(z):
    return norm.ppf(z, loc=0, scale=1)


def check_positivity_or_throw(*args):
    for arg in args:
        if arg <= 0:
            raise ValueError('Value must be positive, however %f is not positive' % arg)


def check_nonnegativity_or_throw(*args):
    for arg in args:
        if arg < 0:
            raise ValueError('Value must be non-negative, however %f is negative' % arg)


def warn_if_normal_approximation_invalid(*args):
    for arg in args:
        if arg <= 5:
            warnings.warn('At least one cell count is less than 5, therefore confidence '
                          'interval approximation is invalid', UserWarning)
            # just print once
            break


def risk_ci(events, total, alpha=0.05, confint='wald'):
    r"""Calculate two-sided risk confidence intervals

    Risk is calculated from

    .. math::

        R = \frac{a}{a+b}

    Wald standard error is

    .. math::

        SE_{Wald} = \left(\frac{1}{a} - \frac{1}{b}\right)^{\frac{1}{2}}

    Hypergeometric standard error is

    .. math::

        SE_{HypGeo} = \left(\frac{a b}{(a+b)^2  (a+b-1)}\right)^{\frac{1}{2}}

    Parameters
    ----------
    events : integer, float
        Number of events/outcomes that occurred
    total : integer, float
        Total number of subjects that could have experienced the event
    alpha : float, optional
        Alpha level. Default is 0.05
    confint : string, optional
        Type of confidence interval to generate. Current options include Wald or Hypergeometric confidence intervals

    Returns
    ---------
    tuple
        Tuple containing risk, lower CL, upper CL, SE

    Note
    ----
    Confidence intervals rely on the central limit theorem, so there must be at least 5 events and 5 nonevents

    Examples
    --------
    Estimate the risk, standard error, and confidence intervals

    >>> from zepid.calc import risk_ci
    >>> r = risk_ci(45, 100)

    Extracting the estimated risk

    >>> r.point_estimate

    Extracting the lower and upper confidence intervals, respectively

    >>> r.lower_bound
    >>> r.upper_bound

    Extracting the standard error

    >>> r.standard_error
    """
    risk = events / total
    c = 1 - alpha / 2
    zalpha = normal_ppf(c)
    if confint == 'wald':
        sd = np.sqrt((risk * (1 - risk)) / total)
        # follows SAS9.4: http://support.sas.com/documentation/cdl/en/procstat/67528/HTML/default/viewer.htm#procstat_
        # freq_details37.htm#procstat.freq.freqbincl
        lower = risk - zalpha * sd
        upper = risk + zalpha * sd
    elif confint == 'hypergeometric':
        sd = np.sqrt(events * (total - events) / (total ** 2 * (total - 1)))
        lower = risk - zalpha * sd
        upper = risk + zalpha * sd
    else:
        raise ValueError('Only wald and hypergeometric confidence intervals are currently supported')
    return Results(risk, lower, upper, sd, alpha, 'risk')


def incidence_rate_ci(events, time, alpha=0.05):
    r"""Calculate two-sided incidence rate confidence intervals. Only Wald-type confidence intervals are currently
    implemented.

    Incidence rate is calculated from

    .. math::

        I = \frac{a}{t}

    Incidence rate standard error is

    .. math::

        SE = \left(\frac{a}{t^2}\right)^\frac{1}{2})

    Parameters
    -------------
    events : integer, float
        Number of events/outcomes that occurred
    time : integer, float
        Total person-time contributed in this group
    alpha : float, optional
        Alpha level. Default is 0.05


    Returns
    ------------
    tuple
        Tuple containing incidence rate, lower CL, upper CL, SE

    Examples
    --------
    Estimate the incidence rate, standard error, and confidence intervals

    >>> from zepid.calc import incidence_rate_ci
    >>> i = incidence_rate_ci(56, 503)

    Extracting the estimated incidence rate

    >>> i.point_estimate

    Extracting the lower and upper confidence intervals, respectively

    >>> i.lower_bound
    >>> i.upper_bound

    Extracting the standard error

    >>> i.standard_error
    """
    c = 1 - alpha / 2
    ir = events / time
    zalpha = normal_ppf(c)
    sd = np.sqrt(events / (time ** 2))
    # https://www.researchgate.net/post/How_to_estimate_standard_error_from_incidence_rate_and_population
    # Incidence rate confidence intervals are a mess, with not sources agreeing...
    # http://www.openepi.com/PersonTime1/PersonTime1.htm
    # zEpid uses a normal approximation. There are too many options for CI's...
    lower = ir - zalpha * sd
    upper = ir + zalpha * sd
    return Results(ir, lower, upper, sd, alpha, 'incidence rate')


def risk_ratio(a, b, c, d, alpha=0.05):
    r"""Calculates the risk ratio and confidence intervals from count data.

    Risk ratio is calculated from

    .. math::

        RR = \frac{a}{a + b} / \frac{c}{c + d}

    Risk ratio standard error is

    .. math::

        SE = \left(\frac{1}{a} - \frac{1}{a + b} + \frac{1}{c} - \frac{1}{c + d}\right)^{\frac{1}{2}}

    Parameters
    ------------
    a : integer, float
        Count of exposed individuals with outcome
    b : integer, float
        Count of unexposed individuals with outcome
    c : integer, float
        Count of exposed individuals without outcome
    d : integer, float
        Count of unexposed individuals without outcome
    alpha : float, optional
        Alpha value to calculate two-sided Wald confidence intervals. Default is 95% confidence interval

    Returns
    --------------
    tuple
        Tuple of risk ratio, lower CL, upper CL, SE

    Examples
    --------
    Estimate the risk ratio, standard error, and confidence intervals

    >>> from zepid.calc import risk_ratio
    >>> rr = risk_ratio(45, 55, 21, 79)

    Extracting the estimated risk ratio

    >>> rr.point_estimate

    Extracting the lower and upper confidence intervals, respectively

    >>> rr.lower_bound
    >>> rr.upper_bound

    Extracting the standard error

    >>> rr.standard_error
    """
    check_positivity_or_throw(a, b, c, d)
    warn_if_normal_approximation_invalid(a, b, c, d)

    zalpha = normal_ppf(1 - alpha / 2)
    r1 = a / (a + b)
    r0 = c / (c + d)
    relrisk = r1 / r0
    sd = np.sqrt((1 / a) - (1 / (a + b)) + (1 / c) - (1 / (c + d)))
    lnrr = np.log(relrisk)
    lcl = np.exp(lnrr - (zalpha * sd))
    ucl = np.exp(lnrr + (zalpha * sd))
    return Results(relrisk, lcl, ucl, sd, alpha, 'risk ratio')


def risk_difference(a, b, c, d, alpha=0.05):
    r"""Calculates the risk difference and confidence intervals from count data.

    Risk difference is calculated as

    .. math::

        RD = \frac{a}{a + b} - \frac{c}{c + d}

    Risk difference standard error is calculated as

    .. math::

        SE = \left(\frac{R_1 \times (1 - R_1)}{a+b} + \frac{R_0 \times (1-R_0)}{c+d}\right)^{\frac{1}{2}}

    Parameters
    ------------
    a : integer, float
        Count of exposed individuals with outcome
    b : integer, float
        Count of unexposed individuals with outcome
    c : integer, float
        Count of exposed individuals without outcome
    d : integer, float
        Count of unexposed individuals without outcome
    alpha : float, optional
        Alpha value to calculate two-sided Wald confidence intervals. Default is 95% confidence interval

    Returns
    --------------
    tuple
        Tuple of risk difference, lower CL, upper CL, SE

    Examples
    --------
    Estimate the risk difference, standard error, and confidence intervals

    >>> from zepid.calc import risk_difference
    >>> rd = risk_difference(45, 55, 21, 79)

    Extracting the estimated risk difference

    >>> rd.point_estimate

    Extracting the lower and upper confidence intervals, respectively

    >>> rd.lower_bound
    >>> rd.upper_bound

    Extracting the standard error

    >>> rd.standard_error
    """
    check_positivity_or_throw(a, b, c, d)
    warn_if_normal_approximation_invalid(a, b, c, d)
    zalpha = normal_ppf(1 - alpha / 2)
    r1 = a / (a + b)
    r0 = c / (c + d)
    riskdiff = r1 - r0
    sd = np.sqrt((r1 * (1 - r1)) / (a + b) + (r0 * (1 - r0)) / (c + d))
    # TODO hypergeometric CL for later implementation
    # sd = np.sqrt(((a * b) / ((a + b) ** 2 * (a + b - 1))) + ((c * d) / (((c + d) ** 2) * (c + d - 1))))
    lcl = riskdiff - (zalpha * sd)
    ucl = riskdiff + (zalpha * sd)
    return Results(riskdiff, lcl, ucl, sd, alpha, 'risk difference')


def number_needed_to_treat(a, b, c, d, alpha=0.05):
    r"""Calculates the number needed to treat and confidence intervals from count data.

    Number needed to treat is calculated as

    .. math::

        NNT = \frac{1}{RD} = \frac{1}{\frac{a}{a + b} - \frac{c}{c + d}}

    Confidence intervals are calculated by taking the inverse of the lower and upper confidence limits of the risk
    difference. The formula for the risk difference standard error is

    .. math::

        SE = \left(\frac{R_1 \times (1 - R_1)}{a+b} + \frac{R_0 \times (1-R_0)}{c+d}\right)^{\frac{1}{2}}

    Note
    ----
    If the risk difference confidence limits cover the null (RD=0), that means the interpretation will switch from
    NNT to NNH (number needed to harm). See Altman 1998 for further details on interpretation is this scenario

    Parameters
    ------------
    a : integer, float
        Count of exposed individuals with outcome
    b : integer, float
        Count of unexposed individuals with outcome
    c : integer, float
        Count of exposed individuals without outcome
    d : integer, float
        Count of unexposed individuals without outcome
    alpha : float, optional
        Alpha value to calculate two-sided Wald confidence intervals. Default is 95% confidence interval

    Returns
    --------------
    tuple
        Tuple of NNT, lower CL, upper CL, SE

    Examples
    --------
    Estimate the number needed to treat, standard error, and confidence intervals

    >>> from zepid.calc import number_needed_to_treat
    >>> nnt = number_needed_to_treat(45, 55, 21, 79)

    Extracting the estimated number needed to treat

    >>> nnt.point_estimate

    Extracting the lower and upper confidence intervals, respectively

    >>> nnt.lower_bound
    >>> nnt.upper_bound

    Extracting the standard error

    >>> nnt.standard_error
    """
    check_positivity_or_throw(a, b, c, d)
    warn_if_normal_approximation_invalid(a, b, c, d)

    zalpha = normal_ppf(1 - alpha / 2)
    r1 = a / (a + b)
    r0 = c / (c + d)
    riskdiff = r1 - r0
    sd = np.sqrt((r1 * (1 - r1)) / (a + b) + (r0 * (1 - r0)) / (c + d))
    # TODO hypergeometric CL for later implementation
    # sd = np.sqrt(((a * b) / ((a + b) ** 2 * (a + b - 1))) + ((c * d) / (((c + d) ** 2) * (c + d - 1))))
    lcl_rd = riskdiff - (zalpha * sd)
    ucl_rd = riskdiff + (zalpha * sd)
    if riskdiff != 0:
        numbnt = 1 / riskdiff
    else:
        numbnt = np.inf
    if lcl_rd != 0:
        lcl = 1 / lcl_rd
    else:
        lcl = np.inf
    if ucl_rd != 0:
        ucl = 1 / ucl_rd
    else:
        ucl = np.inf
    return Results(numbnt, lcl, ucl, sd, alpha, 'number needed to treat')


def odds_ratio(a, b, c, d, alpha=0.05):
    r"""Calculates the odds ratio and confidence interval from count data

    Odds ratio is calculated from

    .. math::

        OR = \frac{a}{b} / \frac{c}{d}

    Odds ratio standard error is

    .. math::

        SE = \left(\frac{1}{a} + \frac{1}{b} + \frac{1}{c} + \frac{1}{d}\right)^{\frac{1}{2}}

    Parameters
    ------------
    a : integer, float
        Count of exposed individuals with outcome
    b : integer, float
        Count of unexposed individuals with outcome
    c : integer, float
        Count of exposed individuals without outcome
    d : integer, float
        Count of unexposed individuals without outcome
    alpha : float, optional
        Alpha value to calculate two-sided Wald confidence intervals. Default is 95% confidence interval

    Returns
    --------------
    tuple
        Tuple of OR, lower CL, upper CL, SE

    Examples
    --------
    Estimate the odds ratio, standard error, and confidence intervals

    >>> from zepid.calc import odds_ratio
    >>> odr = odds_ratio(45, 55, 21, 79)

    Extracting the estimated odds ratio

    >>> odr.point_estimate

    Extracting the lower and upper confidence intervals, respectively

    >>> odr.lower_bound
    >>> odr.upper_bound

    Extracting the standard error

    >>> odr.standard_error
    """
    check_positivity_or_throw(a, b, c, d)
    warn_if_normal_approximation_invalid(a, b, c, d)

    zalpha = normal_ppf(1 - alpha / 2)
    or1 = a / b
    or0 = c / d
    oddsr = or1 / or0
    sd = np.sqrt((1 / a) + (1 / b) + (1 / c) + (1 / d))
    lnor = np.log(oddsr)
    lcl = np.exp(lnor - (zalpha * sd))
    ucl = np.exp(lnor + (zalpha * sd))
    return Results(oddsr, lcl, ucl, sd, alpha, 'odds ratio')


def incidence_rate_ratio(a, c, t1, t2, alpha=0.05):
    r"""Calculates the incidence rate ratio and confidence intervals from count data

    Incidence rate ratio is calculated from

    .. math::

        IR = \frac{a}{t_1} / \frac{c}{t_2}

    Incidence rate ratio standard error is

    .. math::

        SE = \left(\frac{1}{a} + \frac{1}{c}\right)^{\frac{1}{2}}

    Parameters
    ------------
    a : integer, float
        Count of exposed individuals with outcome
    c : integer, float
        Count of unexposed individuals with outcome
    t1 : integer, float
        Person-time contributed by those who were exposed
    t2 : integer, float
        Person-time contributed by those who were unexposed
    alpha : float, optional
        Alpha value to calculate two-sided Wald confidence intervals. Default is 95% confidence interval

    Returns
    --------------
    tuple
        Tuple of IR, lower CL, upper CL, SE

    Examples
    --------
    Estimate the incidence rate ratio, standard error, and confidence intervals

    >>> from zepid.calc import incidence_rate_ratio
    >>> ir = incidence_rate_ratio(45, 21, 109, 158)

    Extracting the estimated incidence rate ratio

    >>> ir.point_estimate

    Extracting the lower and upper confidence intervals, respectively

    >>> ir.lower_bound
    >>> ir.upper_bound

    Extracting the standard error

    >>> ir.standard_error
    """
    check_positivity_or_throw(a, c)
    check_nonnegativity_or_throw(t2, t1)
    warn_if_normal_approximation_invalid(a, c)

    zalpha = normal_ppf(1 - alpha / 2)
    irate1 = a / t1
    irate2 = c / t2
    irater = irate1 / irate2
    sd = np.sqrt((1 / a) + (1 / c))
    lnirr = np.log(irater)
    lcl = np.exp(lnirr - (zalpha * sd))
    ucl = np.exp(lnirr + (zalpha * sd))
    return Results(irater, lcl, ucl, sd, alpha, 'incidence rate ratio')


def incidence_rate_difference(a, c, t1, t2, alpha=0.05):
    r"""Calculates the incidence rate difference and confidence intervals from count data

    Incidence rate difference is calculated from

    .. math::

        ID = \frac{a}{t_1} - \frac{c}{t_2}

    Incidence rate difference standard error is

    .. math::

        SE = \left(\frac{a}{t_1^2} + \frac{c}{t_2^2}\right)^{\frac{1}{2}}

    Parameters
    ------------
    a : integer, float
        Count of exposed individuals with outcome
    c : integer, float
        Count of unexposed individuals with outcome
    t1 : integer, float
        Person-time contributed by those who were exposed
    t2 : integer, float
        Person-time contributed by those who were unexposed
    alpha : float, optional
        Alpha value to calculate two-sided Wald confidence intervals. Default is 95% confidence interval

    Returns
    --------------
    tuple
        Tuple of ID, lower CL, upper CL, SE)

    Examples
    --------
    Estimate the incidence rate ratio, standard error, and confidence intervals

    >>> from zepid.calc import incidence_rate_difference
    >>> ird = incidence_rate_difference(45, 21, 109, 158)

    Extracting the estimated incidence rate ratio

    >>> ird.point_estimate

    Extracting the lower and upper confidence intervals, respectively

    >>> ird.lower_bound
    >>> ird.upper_bound

    Extracting the standard error

    >>> ird.standard_error
    """
    check_positivity_or_throw(a, c)
    check_nonnegativity_or_throw(t2, t1)
    warn_if_normal_approximation_invalid(a, c)

    zalpha = normal_ppf(1 - alpha / 2)
    rated1 = a / t1
    rated2 = c / t2
    irated = rated1 - rated2
    sd = np.sqrt((a / (t1 ** 2)) + (c / (t2 ** 2)))
    lcl = irated - (zalpha * sd)
    ucl = irated + (zalpha * sd)
    return Results(irated, lcl, ucl, sd, alpha, 'incidence rate difference')


def attributable_community_risk(a, b, c, d):
    r"""Calculates the estimated attributable community risk (ACR) from count data. ACR is also known as Population
    Attributable Risk. Since this is commonly confused with the population attributable fraction, the name ACR is used
    to clarify differences in the formulas

    Attributable community risk is calculated as

    .. math::

        ACR = \frac{a + c}{a + b + c + d} - \frac{c}{c + d} = R - R_0

    Parameters
    ------------
    a : integer, float
        Count of exposed individuals with outcome
    b : integer, float
        Count of unexposed individuals with outcome
    c : integer, float
        Count of exposed individuals without outcome
    d : integer, float
        Count of unexposed individuals without outcome

    Returns
    --------------
    float
        Attributable community risk

    Examples
    --------
    Return the attributable community risk

    >>> from zepid.calc import attributable_community_risk
    >>> attributable_community_risk(45, 55, 21, 79)
    """
    check_positivity_or_throw(a, b, c, d)

    rt = (a + c) / (a + b + c + d)
    r0 = c / (c + d)
    return rt - r0


def population_attributable_fraction(a, b, c, d):
    r"""Calculates the population attributable fraction (PAF) from count data

    Population attributable fraction is calculated as

    .. math::

        PAF = \left(\frac{a + c}{a + b + c + d} - \frac{c}{c + d}\right) / \frac{a + c}{a + b + c + d} = (R - R_0) / R

    Parameters
    ------------
    a : integer, float
        Count of exposed individuals with outcome
    b : integer, float
        Count of unexposed individuals with outcome
    c : integer, float
        Count of exposed individuals without outcome
    d : integer, float
        Count of unexposed individuals without outcome

    Returns
    --------------
    float
        Population attributable fraction

    Examples
    --------
    Return the population attributable fraction

    >>> from zepid.calc import population_attributable_fraction
    >>> population_attributable_fraction(45, 55, 21, 79)
    """
    check_positivity_or_throw(a, b, c, d)

    rt = (a + c) / (a + b + c + d)
    r0 = c / (c + d)
    return (rt - r0) / rt


def probability_to_odds(prob):
    r"""Converts given probability (proportion) to odds

    Probability is converted to odds using

    .. math::

        O = \frac{\Pr}{1 - \Pr}

    Parameters
    ---------------
    prob : float, NumPy array
        Probability or array of probabilities to transform into odds

    Returns
    ----------
    odds
        Float or array of odds

    Examples
    --------
    Convert a single probability to an odds

    >>> from zepid.calc import probability_to_odds
    >>> probability_to_odds(0.3)

    Convert an array of probabilities to odds

    >>> import numpy as np
    >>> probs = np.array([0.3, 0.1, 0.2, 0.5, 0.01])
    >>> probability_to_odds(probs)
    """
    return prob / (1 - prob)


def odds_to_probability(odds):
    r"""Converts given odds to probability (proportion)

    Probability is converted to odds using

    .. math::

        \Pr = \frac{O}{1 + O}

    Parameters
    ---------------
    odds : float, NumPy array
        Odds or array of odds to transform into probabilities

    Returns
    ----------
    prob
        Float or array of probabilities

    Examples
    --------
    Convert a single odds to a probability

    >>> from zepid.calc import odds_to_probability
    >>> odds_to_probability(0.45)

    Convert an array of odds to probabilities

    >>> import numpy as np
    >>> odds = np.array([0.3, 0.5, 1, 3.1, 1.1])
    >>> odds_to_probability(odds)
    """
    return odds / (1 + odds)


def logit(prob):
    """Logit transformation of probabilities. Input can be a single probability of array of probabilities

    Parameters
    ----------
    prob : float, array
        A single probability or an array of probabilities

    Returns
    -------
    logit-transformed probabilities
    """
    return np.log(prob / (1 - prob))


def inverse_logit(logodds):
    """Inverse logit transformation. Returns probabilities

    Parameters
    ----------
    logodds : float, array
        A single log-odd or an array of log-odds

    Returns
    -------
    inverse-logit transformed results (i.e. probabilities for log-odds)
    """
    return 1 / (1 + np.exp(-logodds))


def counternull_pvalue(estimate, lcl, ucl, sided='two', alpha=0.05, decimal=3):
    r"""Calculates the counternull p-value. It is useful to prevent over-interpretation of results

    Parameters
    -------------
    estimate : float
        Point estimate for result
    lcl : float
        Lower confidence limit
    ucl : float
        Upper confidence limit
    sided : string, optional
        Whether to compute the upper one-sided, lower one-sided, or two-sided counternull
        p-value. Default is the two-sided
            * 'upper'     Upper one-sided p-value
            * 'lower'     Lower one-sided p-value
            * 'two'       Two-sided p-value
    alpha : float, optional
        Alpha level for p-value. Default is 0.05. Verify that this is the same alpha used to generate confidence
        intervals
    decimal : integer, optional
        Number of decimal places to display. Default is three

    Returns
    ----------
    None
        Function does not return an object. It prints results to the console

    Notes
    --------
    Make sure that the confidence interval points put into the equation match the alpha level calculation

    Examples
    --------
    Calculate the counternull p-value for a single estimate and confidence interval

    >>> from zepid.calc import counternull_pvalue
    >>> counternull_pvalue(-0.1, -0.3, 0.1)

    References
    ----------
    Rosenthal R, Rubin DB. (1994). The counternull value of an effect size: A new statistic. Psychological Science,
    5(6), 329-334.
    """
    zalpha = normal_ppf(1 - alpha / 2)
    se = (ucl - lcl) / (zalpha * 2)
    cnull = 2 * estimate
    up_cn = norm.cdf(x=cnull, loc=estimate, scale=se)
    lp_cn = 1 - up_cn
    lowerp = norm.cdf(x=estimate, loc=cnull, scale=se)
    upperp = 1 - lowerp
    twosip = 2 * min([up_cn, lp_cn])
    print('----------------------------------------------------------------------')
    print('Alpha = ', alpha)
    print('----------------------------------------------------------------------')
    print('Counternull estimate = ', cnull)
    if sided == 'upper':
        print('Upper one-sided counternull p-value: ', round(upperp, decimal))
    elif sided == 'lower':
        print('Lower one-sided counternull p-value: ', round(lowerp, decimal))
    else:
        print('Two-sided counternull p-value: ', round(twosip, decimal))
    print('----------------------------------------------------------------------\n')


def semibayes(prior_mean, prior_lcl, prior_ucl, mean, lcl, ucl, ln_transform=False, alpha=0.05,
              decimal=3, print_results=True):
    r"""A simple Bayesian Analysis. Note that this analysis assumes a normal distribution for the
    continuous measure. See chapter 18 of Modern Epidemiology 3rd Edition (specifically pages 334, 340 for this
    procedure)

    The posterior estimate is calculated as

    .. math::

        E_{posterior} = \frac{\left(E_{prior} \times \frac{1}{Var_{prior}}\right) +
        (E \times \frac{1}{Var})}{E_{prior} \times \frac{1}{Var_{prior}}}

    and the posterior variance is

    .. math::

        Var_{posterior} = \frac{1}{\frac{1}{Var_{prior}} + \frac{1}{Var}}

    Parameters
    ------------
    prior_mean : float
        Prior designated point estimate
    prior_lcl : float
        Prior designated lower confidence limit
    prior_ucl : float
        Prior designated upper confidence limit
    mean : float
        Point estimate result obtained from analysis
    lcl : float
        Lower confidence limit estimate obtained from analysis
    ucl : float
        Upper confidence limit estimate obtained from analysis
    ln_transform : bool, optional
        Whether to natural log transform results before conducting analysis. Should be used for
        RR, OR, or or other Ratio measure. Default is False (use for RD and other absolute measures)
    alpha : float, optional
        Alpha level for confidence intervals. Default is 0.05
    decimal : float, optional
        Number of decimal places to display. Default is three
    print_results : bool, optional
        Whether to print the results of the semi-Bayesian calculations. Default is True

    Returns
    ---------
    tuple
        Tuple of posterior mean, posterior lower CL, posterior upper CL

    Note
    ----
    Make sure that the alpha used to generate the confidence intervals matches the alpha used in this calculation.
    Additionally, this calculation can only handle normally distributed priors and observed

    Examples
    --------
    Posterior Risk Difference

    >>> from zepid.calc import semibayes
    >>> semibayes(prior_mean=-0.15, prior_lcl=-0.5, prior_ucl=0.2, mean=-0.1, lcl=-0.3, ucl=0.1, print_results=False)

    Posterior Risk Ratio

    >>> semibayes(prior_mean=0.9, prior_lcl=0.75, prior_ucl=1.2, mean=0.85, lcl=0.77, ucl=0.91, ln_transform=True)

    References
    ----------
    Rothman KJ, Greenland S, Lash TL. (2008). Modern epidemiology (Vol. 3). Philadelphia: Wolters Kluwer
    Health/Lippincott Williams & Wilkins.
    """
    # Transforming to log scale if ratio measure
    if ln_transform:
        prior_mean = np.log(prior_mean)
        prior_lcl = np.log(prior_lcl)
        prior_ucl = np.log(prior_ucl)
        mean = np.log(mean)
        lcl = np.log(lcl)
        ucl = np.log(ucl)
    zalpha = normal_ppf(1 - alpha / 2)

    # Extracting prior SD
    prior_sd = (prior_ucl - prior_lcl) / (2 * zalpha)
    prior_var = prior_sd ** 2
    prior_w = 1 / prior_var

    # Extracting observed SD
    sd = (ucl - lcl) / (2 * zalpha)
    var = sd ** 2
    w = 1 / var

    # Checking Prior
    check = (mean - prior_mean) / ((var + prior_var) ** (1 / 2))  # ME3 pg340
    approx_check = norm.sf(abs(check)) * 2
    if approx_check < 0.05:
        warnings.warn('The approximate homogeneity check between the prior and the data indicates the prior may be  '
                      'incompatible with the data. Therefore, it may be misleading to summarize the results via a '
                      'semi-Bayesian approach. Either the prior, data, or both are misleading.')

    # Calculating posterior
    post_mean = ((prior_mean * prior_w) + (mean * w)) / (prior_w + w)
    post_var = 1 / (prior_w + w)
    sd = np.sqrt(post_var)
    post_lcl = post_mean - zalpha * sd
    post_ucl = post_mean + zalpha * sd

    # Transforming back if ratio measure
    if ln_transform:
        post_mean = np.exp(post_mean)
        post_lcl = np.exp(post_lcl)
        post_ucl = np.exp(post_ucl)
        prior_mean = np.exp(prior_mean)
        prior_lcl = np.exp(prior_lcl)
        prior_ucl = np.exp(prior_ucl)
        mean = np.exp(mean)
        lcl = np.exp(lcl)
        ucl = np.exp(ucl)

    # Presenting Results
    if print_results:
        print('----------------------------------------------------------------------')
        print('Prior-Data Compatibility Check: p =', round(approx_check, 5))
        print('For compatibility check, small p-values indicate misleading results')
        print('----------------------------------------------------------------------')
        print('Prior Estimate: ', round(prior_mean, decimal))
        print(str(round((1 - alpha) * 100, 1)) + '% Prior Confidence Interval: (', round(prior_lcl, decimal), ', ',
              round(prior_ucl, decimal), ')')
        print('----------------------------------------------------------------------')
        print('Data Estimate: ', round(mean, decimal))
        print(str(round((1 - alpha) * 100, 1)) + '% Confidence Interval: (',
              round(lcl, decimal), ', ', round(ucl, decimal), ')')
        print('----------------------------------------------------------------------')
        print('Posterior Estimate: ', round(post_mean, decimal))
        print(str(round((1 - alpha) * 100, 1)) + '% Posterior Probability Interval: (', round(post_lcl, decimal), ', ',
              round(post_ucl, decimal), ')')
        print('----------------------------------------------------------------------\n')
    return post_mean, post_lcl, post_ucl


def sensitivity(detected, cases, alpha=0.05, confint='wald'):
    r"""Calculate the sensitivity from number of detected cases and the number of total true cases.

    Parameters
    ---------------
    detected : integer, float
        Number of true cases detected via testing criteria
    cases : integer, float
        Total number of true/actual cases
    alpha : float, optional
        Alpha value to calculate two-sided Wald confidence intervals. Default is 95% confidence interval
    confint : string, optional
        Type of confidence interval to generate. Current options include Wald or Hypergeometric confidence intervals


    Returns
    --------
    tuple
        Tuple of sensitivity, lower CL, upper CL, SE

    Examples
    --------
    Calculating sensitivity

    >>> from zepid.calc import sensitivity
    >>> se = sensitivity(90, 100)

    Extract sensitivity

    >>> se[0]

    Extract confidence intervals for sensitivity

    >>> se[1:3]

    Extract standard error

    >>> se[3]
    """
    check_positivity_or_throw(detected, cases)
    warn_if_normal_approximation_invalid(cases)

    if detected > cases:
        raise ValueError('Detected true cases must be less than or equal to the total number of cases')

    sens = detected / cases
    zalpha = norm.ppf(1 - alpha / 2, loc=0, scale=1)
    if confint == 'wald':
        sd = np.sqrt((sens * (1 - sens)) / cases)
        # follows SAS9.4: http://support.sas.com/documentation/cdl/en/procstat/67528/HTML/default/viewer.htm#procstat_
        # freq_details37.htm#procstat.freq.freqbincl
        lower = sens - zalpha * sd
        upper = sens + zalpha * sd
    elif confint == 'hypergeometric':
        sd = np.sqrt(detected * (cases - detected) / (cases ** 2 * (cases - 1)))
        lower = sens - zalpha * sd
        upper = sens + zalpha * sd
    else:
        raise ValueError('Please specify a valid confidence interval')
    return sens, lower, upper, sd


def specificity(detected, noncases, alpha=0.05, confint='wald'):
    """Calculate the specificity from number of false detections and the number of total non-cases.

    Parameters
    ---------------
    detected : integer, float
        Number of false cases detected via testing criteria
    noncases : integer, float
        Total number of non-cases
    alpha : float, optional
        Alpha value to calculate two-sided Wald confidence intervals. Default is 95% confidence interval
    confint : string, optional
        Type of confidence interval to generate. Current options include Wald or Hypergeometric confidence intervals

    Returns
    --------
    tuple
        Tuple of specificity, lower CL, upper CL, SE

    Examples
    --------
    Calculating specificity

    >>> from zepid.calc import specificity
    >>> sp = specificity(88, 100)

    Extract specificity

    >>> sp[0]

    Extract confidence intervals for specificity

    >>> sp[1:3]

    Extract standard error

    >>> sp[3]
    """
    check_positivity_or_throw(detected, noncases)
    warn_if_normal_approximation_invalid(noncases)

    if detected > noncases:
        raise ValueError('Detected true cases must be less than or equal to the total number of cases')
    spec = 1 - (detected / noncases)
    zalpha = norm.ppf(1 - alpha / 2, loc=0, scale=1)
    if confint == 'wald':
        sd = np.sqrt((spec * (1 - spec)) / noncases)
        # follows SAS9.4: http://support.sas.com/documentation/cdl/en/procstat/67528/HTML/default/viewer.htm#procstat_
        # freq_details37.htm#procstat.freq.freqbincl
        lower = spec - zalpha * sd
        upper = spec + zalpha * sd
    elif confint == 'hypergeometric':
        sd = np.sqrt(detected * (noncases - detected) / (noncases ** 2 * (noncases - 1)))
        lower = spec - zalpha * sd
        upper = spec + zalpha * sd
    else:
        raise ValueError('Please specify a valid confidence interval')
    return spec, lower, upper, sd


def ppv_converter(sensitivity, specificity, prevalence):
    r"""Generates the positive predictive value from designated sensitivity, specificity, and prevalence.

    Positive predictive value is calculated using

    .. math::

        PPV = \frac{Se \times P}{Se \times P + (1-Sp) (1-P)}

    Parameters
    -------------
    sensitivity : float
        Sensitivity of the testing criteria
    specificity : float
        Specificity of the testing criteria
    prevalence : float
        Prevalence of the outcome in the population

    Returns
    ------------
    float
        Positive predictive value

    Examples
    --------
    Calculate the positivity predictive value

    >>> from zepid.calc import ppv_converter
    >>> ppv_converter(0.9, 0.88, 0.15)
    """
    if (sensitivity > 1) or (specificity > 1) or (prevalence > 1):
        raise ValueError('sensitivity/specificity/prevalence cannot be greater than 1')
    if (sensitivity < 0) or (specificity < 0) or (prevalence < 0):
        raise ValueError('sensitivity/specificity/prevalence cannot be less than 0')
    sens_prev = sensitivity * prevalence
    nspec_nprev = (1 - specificity) * (1 - prevalence)
    ppv = sens_prev / (sens_prev + nspec_nprev)
    return ppv


def npv_converter(sensitivity, specificity, prevalence):
    r"""Generates the negative predictive value from designated sensitivity, specificity, and prevalence.

    .. math::

        NPV = \frac{Sp \times (1-P)}{(1-Se) \times P + Sp \times (1-P)}

    Parameters
    -------------
    sensitivity : float
        Sensitivity of the testing criteria
    specificity : float
        Specificity of the testing criteria
    prevalence : float
        Prevalence of the outcome in the population

    Returns
    ------------
    float
        Negative predictive value

    Examples
    --------
    Calculate the positivity predictive value

    >>> from zepid.calc import npv_converter
    >>> npv_converter(0.9, 0.88, 0.15)
    """
    if (sensitivity > 1) or (specificity > 1) or (prevalence > 1):
        raise ValueError('sensitivity/specificity/prevalence cannot be greater than 1')
    if (sensitivity < 0) or (specificity < 0) or (prevalence < 0):
        raise ValueError('sensitivity/specificity/prevalence cannot be less than 0')
    spec_nprev = specificity * (1 - prevalence)
    nsens_prev = (1 - sensitivity) * prevalence
    npv = spec_nprev / (spec_nprev + nsens_prev)
    return npv


def screening_cost_analyzer(cost_miss_case, cost_false_pos, prevalence, sensitivity,
                            specificity, population=10000, decimal=3):
    """Compares the cost of sensivitiy/specificity of screening criteria to treating the entire population
    as test-negative and test-positive. The lowest per capita cost is considered the ideal choice. Note that
    this function only provides relative costs

    Parameters
    ----------------
    cost_miss_case : float
        The relative cost of missing a case, compared to false positives. In general, set this to 1 then change the
        value under 'cost_false_pos' to reflect the relative cost
    cost_false_pos : float
        The relative cost of a false positive case, compared to a missed case
    prevalence : float
        The prevalence of the disease in the population. Must be a float
    sensitivity : float
        The sensitivity level of the screening test. Must be a float
    specificity : float
        The specificity level of the screening test. Must be a float
    population : float
        The population size to set. Choose a larger value since this is only necessary for total calculations. Default
        is 10,000
    decimal : integer
        Amount of decimal points to display. Default value is 3

    Returns
    ---------
    None
        Prints results to console

    Note
    ----
    When calculating costs, be sure to consult experts in health policy or related fields.  Costs should encompass more
    than just monetary costs, like relative costs (regret, disappointment, stigma, disutility, etc.). Careful
    consideration of relative costs between false positive and false negatives needs to be considered.

    Examples
    --------
    Calculate the (relative) cost for the proposed screening strategy

    >>> from zepid.calc import screening_cost_analyzer
    >>> screening_cost_analyzer(cost_miss_case=1, cost_false_pos=3, prevalence=0.15, sensitivity=0.9, specificity=0.88)
    """
    warnings.warn("NOTE: When calculating costs, be sure to consult experts in health policy or related fields."
                  " Costs should encompass more than only monetary costs, like relative costs (regret, "
                  "disappointment, stigma, disutility, etc.)", UserWarning)

    if (sensitivity > 1) | (specificity > 1):
        raise ValueError('sensitivity/specificity/prevalence cannot be greater than 1')
    disease = population * prevalence
    disease_free = population - disease

    # TEST: no positives
    nt_cost = disease * cost_miss_case
    pc_nt_cost = nt_cost / population

    # TEST: all postives
    t_cost = disease_free * cost_false_pos
    pc_t_cost = t_cost / population

    # TEST: criteria
    cost_b = disease - (disease * sensitivity)
    cost_c = disease_free - (disease_free * specificity)
    ct_cost = (cost_miss_case * cost_b) + (cost_false_pos * cost_c)
    pc_ct_cost = ct_cost / population

    # Present results
    print('----------------------------------------------------------------------')
    print('Treat everyone as Test-Negative')
    print('Total relative cost:\t\t', round(nt_cost, decimal))
    print('Per Capita relative cost:\t', round(pc_nt_cost, decimal))
    print('----------------------------------------------------------------------')
    print('Treat everyone as Test-Positive')
    print('Total relative cost:\t\t', round(t_cost, decimal))
    print('Per Capita relative cost:\t', round(pc_t_cost, decimal))
    print('----------------------------------------------------------------------')
    print('Treating by Screening Test')
    print('Total relative cost:\t\t', round(ct_cost, decimal))
    print('Per Capita relative cost:\t', round(pc_ct_cost, decimal))
    print('----------------------------------------------------------------------')
    if pc_ct_cost > pc_nt_cost:
        print('Screening program is more costly than treating everyone as a test-negative')
    if pc_nt_cost > pc_ct_cost > pc_t_cost:
        print('Screening program is cost efficient')
    if (pc_t_cost < pc_ct_cost) and (pc_t_cost < pc_nt_cost):
        print('Treating everyone as test-positive is least costly')
    print('----------------------------------------------------------------------\n')


def rubins_rules(point_estimates, std_error):
    r"""Function to merge multiple imputed data sets into a single summary estimate and variance. Results are based on
    Rubin's Rules for merging estimates. The summary point estimate is calculated via

    .. math::

        \bar{\beta} = m^{-1} \sum_{k=1}^m \hat{\beta_k}

    where m is the number of imputed data sets. The variance is calculated via

    .. math::

        Var(\hat{\beta}) = m_{-1} \sum_{k=1}^m Var(\hat{\beta}) + (1 + m^{-1})(m-1)^{-1} \sum_{k=1}^m
        (\hat{\beta_k} - \bar{\beta})^2

    The variance is constructed from the within-sample variance and the between sample variance

    Notes
    -----
    If your point estimates correspond to ratios, be sure to provide the natural-log transformed point estimates and
    the variance of the natural-log estimate

    Parameters
    ----------
    point_estimates : list
        Container object of the point estimates
    std_error : list
        Container object of the estimate standard errors

    Returns
    -------
    tuple
        Tuple of summary beta, summary standard error

    Examples
    --------
    >>> from zepid.calc import rubins_rules
    >>> rr_est = []
    >>> rr_std = []

    Calculating summary estimate

    >>> b = rubins_rules(rr_est, rr_std)

    Printing the summary risk ratio

    >>> print("RR = ", np.exp(b[0]))
    >>> print("95% LCL:", np.exp(b[0] - 1.96*b[1]))
    >>> print("95% UCL:", np.exp(b[0] + 1.96*b[1]))

    References
    ----------
    Rubin DB. (2004). Multiple imputation for nonresponse in surveys (Vol. 81). John Wiley & Sons.
    """
    if len(point_estimates) != len(std_error):
        raise ValueError("The number of point estimates and variances must be the same")

    # Calculate summary point estimate
    beta = np.mean(point_estimates)

    # Calculate summary variance estimate
    variance = np.array(std_error)**2
    var_between = np.sum((np.array(point_estimates) - beta)**2)
    var_within = np.mean(variance)
    var = var_within + (1 + len(variance)**-1)*((len(variance) - 1)**-1)*var_between

    return beta, np.sqrt(var)


def s_value(pvalue):
    r"""Function to calculate the S-value. The 'S' stands for Shannon information or surprisal values. The name comes
    from Claude Shannon for this work to information theory. S-values are calculated from p-values using the
    following transformation

    .. math::

        s = -\log(p)

    The S-value transformation allows a more intuitive explanation of what p-values tell us about the null hypothesis
    and alternative hypothesis compatibility. The S-value tells us how many 'bits' of information exist against the null
    hypothesis. For an example, a S-value of 5.1 is no more surprising than seeing heads for 5 fair coin tosses. The
    S-value should be rounded down in the interpretation

    Note
    ----
    S-values do NOT have a significant cut-point. Rather this transformation is to help build intuition what information
    a p-values is providing and the corresponding 'surprisal' of a result

    Parameters
    ----------
    pvalue : float, container
        P-value (or array of p-values) to convert into a S-value(s)

    Returns
    -------
    array
        NumPy array of calculated S-values

    Examples
    --------
    >>> from zepid.calc import s_value
    >>> s_value(pvalue=0.05)

    References
    ----------
    Greenland S. (2019). Valid P-values behave exactly as they should: Some misleading criticisms of P-values and their
    resolution with S-values. The American Statistician, 73(sup1), 106-114.

    Amrhein V, Trafimow D, & Greenland S. (2018). Inferential Statistics as Descriptive Statistics: There is No
    Replication Crisis if We Don’t Expect Replication. The American Statistician.
    """
    return -1 * np.log2(np.array(pvalue))


def probability_bounds(v, bounds):
    """Function to generate bounded values for probabilities. Specifically this function is used in multiple
    estimators to generate bounded probabilities. This is available for both symmetric and asymmetric bounds.

    Parameters
    ----------
    v : numpy.array
        Array of values to bound
    bounds : float, list, numpy.array
        Bounds to apply to v. If only a single value is provided, then symmetric bounds are used.

    Returns
    -------
    numpy.array of bounded values
    """
    v = np.asarray(v)
    if type(bounds) is float:  # Symmetric Bounding
        if bounds < 0 or bounds > 1:
            raise ValueError('Bound value must be between (0, 1)')
        v[v < bounds] = bounds
        v[v > 1 - bounds] = 1 - bounds

    elif type(bounds) is str:  # Catching string inputs
        raise ValueError('Bounds must either be a float between (0, 1), or a collection of floats between (0, 1)')

    elif type(bounds) is int:  # Catching string inputs
        raise ValueError('Bounds must either be a float between (0, 1), or a collection of floats between (0, 1)')

    else:  # Asymmetric bounds
        if bounds[0] > bounds[1]:
            raise ValueError('Bound thresholds must be listed in ascending order')
        if len(bounds) > 2:
            warnings.warn('It looks like your specified bounds is more than two floats. Only the first two '
                          'specified bounds are used by the bound statement. So only ' +
                          str(bounds[0:2]) + ' will be used', UserWarning)
        if type(bounds[0]) is str or type(bounds[1]) is str:
            raise ValueError('Bounds must be floats between (0, 1)')
        if (bounds[0] < 0 or bounds[1] > 1) or (bounds[0] < 0 or bounds[1] > 1):
            raise ValueError('Both bound values must be between (0, 1)')
        v[v < bounds[0]] = bounds[0]
        v[v > bounds[1]] = bounds[1]

    return v



import numpy as np

# Utilities only meant for the doubly-robust branch


def tmle_unit_bounds(y, mini, maxi, bound):
    # bounding for continuous outcomes
    v = (y - mini) / (maxi - mini)
    v = np.where(np.less(v, bound), bound, v)
    v = np.where(np.greater(v, 1-bound), 1-bound, v)
    return v


def tmle_unit_unbound(ystar, mini, maxi):
    # unbounding of bounded continuous outcomes
    return ystar*(maxi - mini) + mini



import warnings
import numpy as np
import pandas as pd
from scipy.stats import norm
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.genmod.families import links
from statsmodels.nonparametric.smoothers_lowess import lowess
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.ticker as mticker


class EffectMeasurePlot:
    """Used to generate effect measure (AKA forest) plots. Estimates and confidence intervals are plotted in a diagram
    on the left and a table of the corresponding estimates is provided in the same plot. See the Graphics page on
    ReadTheDocs examples of the plots

    Parameters
    --------------
    label : list
        List of labels to use for y-axis
    effect_measure : list
        List of numbers for point estimates to plot. If point estimate has trailing zeroes,
        input as a character object rather than a float
    lcl : list
        List of numbers for upper confidence limits to plot. If point estimate has trailing
        zeroes, input as a character object rather than a float
    ucl : list
        List of numbers for upper confidence limits to plot. If point estimate has
        trailing zeroes, input as a character object rather than a float

    Examples
    -------------
    Setting up the data to plot

    >>> from matplotlib.pyplot as plt
    >>> from zepid.graphics import EffectMeasurePlot
    >>> lab = ['One','Two']
    >>> emm = [1.01,1.31]
    >>> lcl = ['0.90',1.01]  # String allows for trailing zeroes in the table
    >>> ucl = [1.11,1.53]

    Setting up the plot, measure labels, and point colors

    >>> x = EffectMeasurePlot(lab, emm, lcl, ucl)
    >>> x.labels(effectmeasure='RR')  # Changing label of measure
    >>> x.colors(pointcolor='r')  # Changing color of the points

    Generating matplotlib axes object of forest plot

    >>> x.plot(t_adjuster=0.13)
    >>> plt.show()
    """
    def __init__(self, label, effect_measure, lcl, ucl):
        self.df = pd.DataFrame()
        self.df['study'] = label
        self.df['OR'] = effect_measure
        self.df['LCL'] = lcl
        self.df['UCL'] = ucl
        self.df['OR2'] = self.df['OR'].astype(str).astype(float)
        if (all(isinstance(item, float) for item in lcl)) & (all(isinstance(item, float) for item in effect_measure)):
            self.df['LCL_dif'] = self.df['OR'] - self.df['LCL']
        else:
            self.df['LCL_dif'] = (pd.to_numeric(self.df['OR'])) - (pd.to_numeric(self.df['LCL']))
        if (all(isinstance(item, float) for item in ucl)) & (all(isinstance(item, float) for item in effect_measure)):
            self.df['UCL_dif'] = self.df['UCL'] - self.df['OR']
        else:
            self.df['UCL_dif'] = (pd.to_numeric(self.df['UCL'])) - (pd.to_numeric(self.df['OR']))
        self.em = 'OR'
        self.ci = '95% CI'
        self.scale = 'linear'
        self.center = 1
        self.errc = 'dimgrey'
        self.shape = 'd'
        self.pc = 'k'
        self.linec = 'gray'

    def labels(self, **kwargs):
        """Function to change the labels of the outputted table. Additionally, the scale and reference
        value can be changed.

        Parameters
        -------------
        effectmeasure : string, optional
            Changes the effect measure label
        conf_int : string, optional
            Changes the confidence interval label
        scale : string, optional
            Changes the scale to either log or linear
        center : float, integer, optional
            Changes the reference line for the center
        """
        if 'effectmeasure' in kwargs:
            self.em = kwargs['effectmeasure']
        if 'conf_int' in kwargs:
            self.ci = kwargs['conf_int']
        if 'scale' in kwargs:
            self.scale = kwargs['scale']
        if 'center' in kwargs:
            self.center = kwargs['center']

    def colors(self, **kwargs):
        """Function to change colors and shapes.

        Parameters
        ---------------
        errorbarcolor : string, optional
            Changes the error bar colors
        linecolor : string, optional
            Changes the color of the reference line
        pointcolor : string, optional
            Changes the color of the points
        pointshape : string, optional
            Changes the shape of points
        """
        if 'errorbarcolor' in kwargs:
            self.errc = kwargs['errorbarcolor']
        if 'pointshape' in kwargs:
            self.shape = kwargs['pointshape']
        if 'linecolor' in kwargs:
            self.linec = kwargs['linecolor']
        if 'pointcolor' in kwargs:
            self.pc = kwargs['pointcolor']

    def plot(self, figsize=(3, 3), t_adjuster=0.01, decimal=3, size=3, max_value=None, min_value=None):
        """Generates the matplotlib effect measure plot with the default or specified attributes.
        The following variables can be used to further fine-tune the effect measure plot

        Parameters
        -----------------
        figsize : tuple, optional
            Adjust the size of the figure. Syntax is same as matplotlib `figsize`
        t_adjuster : float, optional
            Used to refine alignment of the table with the line graphs. When generate plots, trial and error for this
            value are usually necessary. I haven't come up with an algorithm to determine this yet...
        decimal : integer, optional
            Number of decimal places to display in the table
        size : integer,
            Option to adjust the size of the lines and points in the plot
        max_value : float, optional
            Maximum value of x-axis scale. Default is None, which automatically determines max value
        min_value : float, optional
            Minimum value of x-axis scale. Default is None, which automatically determines min value

        Returns
        ---------
        matplotlib axes
        """
        tval = []
        ytick = []
        for i in range(len(self.df)):
            if not np.isnan(self.df['OR2'][i]):
                if ((isinstance(self.df['OR'][i], float)) & (isinstance(self.df['LCL'][i], float)) &
                        (isinstance(self.df['UCL'][i], float))):
                    tval.append([round(self.df['OR2'][i], decimal), (
                                '(' + str(round(self.df['LCL'][i], decimal)) + ', ' +
                                str(round(self.df['UCL'][i], decimal)) + ')')])
                else:
                    tval.append(
                        [self.df['OR'][i], ('(' + str(self.df['LCL'][i]) + ', ' + str(self.df['UCL'][i]) + ')')])
                ytick.append(i)
            else:
                tval.append([' ', ' '])
                ytick.append(i)
        if max_value is None:
            if pd.to_numeric(self.df['UCL']).max() < 1:
                maxi = round(((pd.to_numeric(self.df['UCL'])).max() + 0.05),
                             2)  # setting x-axis maximum for UCL less than 1
            if (pd.to_numeric(self.df['UCL']).max() < 9) and (pd.to_numeric(self.df['UCL']).max() >= 1):
                maxi = round(((pd.to_numeric(self.df['UCL'])).max() + 1),
                             0)  # setting x-axis maximum for UCL less than 10
            if pd.to_numeric(self.df['UCL']).max() > 9:
                maxi = round(((pd.to_numeric(self.df['UCL'])).max() + 10),
                             0)  # setting x-axis maximum for UCL less than 100
        else:
            maxi = max_value
        if min_value is None:
            if pd.to_numeric(self.df['LCL']).min() > 0:
                mini = round(((pd.to_numeric(self.df['LCL'])).min() - 0.1), 1)  # setting x-axis minimum
            if pd.to_numeric(self.df['LCL']).min() < 0:
                mini = round(((pd.to_numeric(self.df['LCL'])).min() - 0.05), 2)  # setting x-axis minimum
        else:
            mini = min_value
        plt.figure(figsize=figsize)  # blank figure
        gspec = gridspec.GridSpec(1, 6)  # sets up grid
        plot = plt.subplot(gspec[0, 0:4])  # plot of data
        tabl = plt.subplot(gspec[0, 4:])  # table of OR & CI
        plot.set_ylim(-1, (len(self.df)))  # spacing out y-axis properly
        if self.scale == 'log':
            try:
                plot.set_xscale('log')
            except:
                raise ValueError('For the log scale, all values must be positive')
        plot.axvline(self.center, color=self.linec, zorder=1)
        plot.errorbar(self.df.OR2, self.df.index, xerr=[self.df.LCL_dif, self.df.UCL_dif], marker='None', zorder=2,
                      ecolor=self.errc, elinewidth=(size / size), linewidth=0)
        plot.scatter(self.df.OR2, self.df.index, c=self.pc, s=(size * 25), marker=self.shape, zorder=3,
                     edgecolors='None')
        plot.xaxis.set_ticks_position('bottom')
        plot.yaxis.set_ticks_position('left')
        plot.get_xaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter())
        plot.get_xaxis().set_minor_formatter(matplotlib.ticker.NullFormatter())
        plot.set_yticks(ytick)
        plot.set_xlim([mini, maxi])
        plot.set_xticks([mini, self.center, maxi])
        plot.set_xticklabels([mini, self.center, maxi])
        plot.set_yticklabels(self.df.study)
        plot.yaxis.set_ticks_position('none')
        plot.invert_yaxis()  # invert y-axis to align values properly with table
        tb = tabl.table(cellText=tval, cellLoc='center', loc='right', colLabels=[self.em, self.ci],
                        bbox=[0, t_adjuster, 1, 1])
        tabl.axis('off')
        tb.auto_set_font_size(False)
        tb.set_fontsize(12)
        for key, cell in tb.get_celld().items():
            cell.set_linewidth(0)
        return plot


def functional_form_plot(df, outcome, var, f_form=None, outcome_type='binary', discrete=False, link_dist=None,
                         loess=True, loess_value=0.25, legend=True, model_results=True, points=False):
    """Creates a functional form plot to aid in functional form assessment for continuous/discrete variables. Plots can
    be created for binary and continuous outcomes. Default options are set to create a functional form plot for a
    binary outcome. To convert to a continuous outcome, `outcome_type` needs to be changed, in addition to
    the `link_dist`

    Parameters
    ------------
    df : DataFrame
        Pandas dataframe that contains the variables of interest
    outcome : string
        Column name of the outcome variable of interest
    var : string
        Column name of the variable of interest for the functional form assessment
    f_form : string, optional
        Regression equation of the functional form to assess. Default is None, which will produce a linear functional
        form. Input the regression equation following the `patsy` syntax. For example, 'var + var_sq'
    outcome_type : string, optional
        Variable type of the outcome variable. Currently, only binary and continuous variables are supported. Default
        is 'binary'
    link_dist : optional
        Link and distribution for the GLM regression equation. Change this to any valid link and distributions
        supported by `statsmodels`. Default is `None`, which defaults to logistic regression
    loess_value : float, optional
        Fraction of observations to use to fit the LOESS curve. This may need to be changed iteratively to determine
        which percent works best for the data. Default is 0.25
    legend : bool, optional
        Turn the legend on or off. Default is True, displaying the legend in the graph
    model_results : bool, optional
        Whether to produce the model results. Default is True, which provides model results
    loess : bool, optional
        Whether to plot the LOESS curve along with the functional form. Default is True
    points : bool, optional
        Whether to plot the data points, where size is relative to the number of observations. Default is False
    discrete : bool, optional
        If your data is truly continuous, leave setting to auto bin the dat. Will automatically bin observations into
        categories for fitting a model with a disjoint indicator. If data is discrete, you can set this to `True` to
        use the actual values for the disjoint indicator. If you get a perfect `SeparationError` from `statsmodels`, it
        means you might have to reshift your categories.

    Returns
    -----------
    matplotlib axes
        Returns a matplotlib graph with a LOESS line (dashed red-line), regression line (sold blue-line), and
        confidence interval (shaded blue)

    Examples
    ------------
    Setting up the environment

    >>> from zepid import load_sample_data
    >>> from zepid.graphics import functional_form_plot
    >>> import matplotlib.pyplot as plt
    >>> df = load_sample_data(timevary=False)
    >>> df['cd4_sq'] = df['cd4']**2

    Creating a functional form plot for a linear functional form

    >>> functional_form_plot(df, outcome='dead', var='cd4')
    >>> plt.show()

    Functional form assessment for a quadractic functional form

    >>> functional_form_plot(df, outcome='dead', var='cd4', f_form='cd4 + cd4_sq')
    >>> plt.show()

    Varying the LOESS value (increased LOESS value to smooth LOESS curve further)

    >>> functional_form_plot(df, outcome='dead', var='cd4', loess_value=0.5)
    >>> plt.show()

    Removing the LOESS curve and the legend from the plot

    >>> functional_form_plot(df, outcome='dead', var='cd4', loess=False, legend=False)
    >>> plt.show()

    Adding summary points to the plot. Points are grouped together and their size reflects their relative n

    >>> functional_form_plot(df, outcome='dead', var='cd4', loess=False, legend=False, points=True)
    >>> plt.show()

    Functional form assessment for a discrete variable (age)

    >>> functional_form_plot(df, outcome='dead', var='age0', discrete=True)
    >>> plt.show()
    """
    # Copying out the dataframe to a new object we will manipulate a bit
    rf = df.copy()
    rf = rf.dropna(subset=[var, outcome]).sort_values(by=[var, outcome]).reset_index()
    if rf.shape[0] != df.shape[0]:
        warnings.warn('Warning: missing observations of model variables are dropped. ' +
                      str(int(df.shape[0] - rf.shape[0])) +
                      ' observations were dropped from the functional form assessment')

    # Functional form for the model
    if f_form is None:
        f_form = var
    else:
        pass

    # Generating Models
    if outcome_type == 'binary':
        if link_dist is None:
            link_dist = sm.families.family.Binomial()
        else:
            pass
    elif outcome_type == 'continuous':
        if link_dist is None:
            link_dist = sm.families.family.Gaussian(sm.families.links.identity)
        else:
            pass
    else:
        raise ValueError('Only binary or continuous outcomes are currently supported')

    # Generating LOESS or points if requested
    ax = plt.gca()
    if loess or points:
        if outcome_type == 'binary':
            if discrete is False:
                # Binning continuous variable into categories to get "General" functional form
                categories = int((np.max(rf[var]) - np.min(rf[var])) / 5)
                if model_results:
                    print('''A total of ''' + str(categories) + ''' categories were created. If you would like to
                            influence  the number of categories the spline is fit to, do the following\n\tIncrease:
                            multiply by constant >1\n\tDecrease: multiply by contast <1 and >0''')
                rf['vbin'] = pd.qcut(rf[var], q=categories, duplicates='drop').cat.codes
                djm = smf.glm(outcome + '~ C(vbin)', rf, family=link_dist).fit()
            else:
                djm = smf.glm(outcome + '~ C(' + var + ')', rf, family=link_dist).fit()
            djf = djm.get_prediction(rf).summary_frame()
            dj = pd.concat([rf, djf], axis=1)
            dj.sort_values(var, inplace=True)
            if points:
                pf = dj.groupby(by=[var, 'mean']).count().reset_index()
                ax.scatter(pf[var], pf['mean'], s=[100 * (n / np.max(pf[var])) for n in pf[var]],
                           color='gray', label='Data point')
            if loess:
                yl = lowess(list(dj['mean']), list(dj[var]), frac=loess_value)
                lowess_x = list(zip(*yl))[0]
                lowess_y = list(zip(*yl))[1]
                ax.plot(lowess_x, lowess_y, '--', color='red', linewidth=1, label='LOESS')
        if outcome_type == 'continuous':
            if points:
                pf = rf.groupby(by=[var, outcome]).count().reset_index()
                ax.scatter(pf[var], pf[outcome], color='gray', label='Data point')
            if loess:
                yl = lowess(list(rf[outcome]), list(rf[var]), frac=loess_value)
                lowess_x = list(zip(*yl))[0]
                lowess_y = list(zip(*yl))[1]
                ax.plot(lowess_x, lowess_y, '--', color='red', linewidth=1, label='LOESS')

    # Functional form model fitting
    ffm = smf.glm(outcome + ' ~ ' + f_form, rf, family=link_dist).fit()
    if model_results is True:
        print(ffm.summary())
        print('AIC: ', ffm.aic)
        print('BIC: ', ffm.bic)
    fff = ffm.get_prediction(rf).summary_frame()
    ff = pd.concat([rf, fff], axis=1)
    ff.sort_values(var, inplace=True)

    # Generating plot for functional form
    ax.fill_between(ff[var], ff['mean_ci_upper'], ff['mean_ci_lower'], alpha=0.1, color='blue', label='95% CI')
    ax.plot(ff[var], ff['mean'], '-', color='blue', label='Regression')
    ax.set_xlabel(var)
    ax.set_ylabel('Outcome')
    if legend is True:
        ax.legend()
    return ax


def pvalue_plot(point, sd, color='b', fill=True, null=0, alpha=None):
    """Creates a plot of the p-value distribution based on a point estimate and standard deviation.
    I find this plot to be useful to explain p-values and how much evidence weight you have in a
    specific value. I think it is useful to explain what exactly a p-value tells you. Note that this
    plot only works for measures on a linear scale (i.e. it will plot exp(log(RR)) incorrectly). It also
    helps to understand what exactly confidence intervals are telling you. These  plots are based on
    Rothman's Epidemiology 2nd Edition pg 152-153 and explained more fully within.

    Parameters
    -------------
    point : float
        Point estimate. Must be on a linear scale (RD / log(RR))
    sd : float
        Standard error of the estimate. Must for linear scale (SE(RD) / SE(log(RR)))
    color : str, optional
        Change color of p-value plot
    fill : bool, optional
        Hhether to fill the curve under the p-value distribution. Setting to False prevents fill
    null : float, integer, optional
        The main value to compare to. The default is zero
    alpha : float, optional
        Whether to draw a line designating significance level area. Default is None, which does not draw this line.
        Generally, would be set to 0.05 to correspond to the widely used alpha of 0.05

    Returns
    -----------
    matplotlib axes

    Examples
    -----------
    Setting up the environment

    >>> from zepid.graphics import pvalue_plot
    >>> import matplotlib.pyplot as plt

    Basic P-value plot

    >>> pvalue_plot(point=-0.1, sd=0.061, color='r')
    >>> plt.show()

    P-value plot with significance line drawn at 'alpha'

    >>> pvalue_plot(point=-0.1, sd=0.061, color='r', alpha=0.025)
    >>> plt.show()

    P-value plot with different comparison value

    >>> pvalue_plot(point=-0.1, sd=0.061, color='r', null=0.1)
    >>> plt.show()

    References
    ----------
    Rothman KJ. (2012). Epidemiology: an introduction. Oxford university press.
    """
    if point <= null:
        lower = (point - 3 * sd)
        if (point + 3 * sd) < 0:
            upper = point + 3 * sd
        else:
            upper = null + 3 * sd
    if point > null:
        upper = (point + 3 * sd)
        if (point - 3 * sd) > 0:
            lower = null - 3 * sd
        else:
            lower = point - 3 * sd
    ax = plt.gca()
    x1 = np.linspace(lower, point, 100)
    x2 = np.linspace(point, upper, 100)
    ax.plot(x2, 2 * (1 - norm.cdf(x2, loc=point, scale=sd)), c=color)
    ax.plot(x1, 2 * norm.cdf(x1, loc=point, scale=sd), c=color)
    if fill == True:
        ax.fill_between(x2, 2 * (1 - norm.cdf(x2, loc=point, scale=sd)), color=color, alpha=0.2)
        ax.fill_between(x1, 2 * norm.cdf(x1, loc=point, scale=sd), color=color, alpha=0.2)
    ax.vlines(null, 0, 1, colors='k')
    ax.set_xlim([lower, upper])
    ax.set_ylim([0, 1])
    ax.set_ylabel('P-value')
    if alpha is not None:
        ax.hlines(alpha, lower, upper)
    return ax


def spaghetti_plot(df, idvar, variable, time):
    """Create a spaghetti plot by an ID variable. A spaghetti plot can be useful for visualizing
    trends or looking at longitudinal data patterns for individuals all at once.

    Parameters
    ------------
    df : DataFrame
        Pandas dataframe containing variables of interest
    idvar : str
        ID variable for observations. This should indicate the group or individual followed over the time variable
    variable : str
        Variable of interest to see how it varies over time
    time : str
        Time or other variable in which the variable variation occurs

    Returns
    -----------
    matplotlib axes

    Examples
    -----------
    Setting up the environment

    >>> from zepid import load_sample_data
    >>> from zepid.graphics import spaghetti_plot
    >>> df = load_sample_data(timevary=True)

    Generating spaghetti plot for changing CD4 count

    >>> spaghetti_plot(df, idvar='id', variable='cd4', time='enter')
    >>> plt.show()
    """
    ax = plt.gca()
    for i in df[idvar].unique():
        s = df.loc[df[idvar] == i].copy()
        s.sort_values(time, ascending=False)
        ax.plot(s[time], s[variable])
    ax.set_xlabel(time)
    ax.set_ylabel(variable)
    return ax


def roc(df, true, threshold, youden_index=True):
    """Generate a Receiver Operator Curve from true values and predicted probabilities. Youden's Index can also be
    calculated. Youden's index is calculated as

    .. math::

        P_{Yi} = max(Se_i + Sp_i - 1)

    Parameters
    ------------
    df : DataFrame
        Pandas dataframe containing variables of interest
    true : str
        True designation of the outcome (1, 0)
    threshold : str
        Predicted probabilities for the outcome
    youden_index : bool, optional
        Whether to calculate Youden's index. Youden's index maximizes both sensitivity and specificity. The formula
        finds the maximum of (sensitivity + specificity - 1)

    Returns
    -----------
    matplotlib axes

    Examples
    --------
    Creating a dataframe with true disease status (`'d'`) and predicted probability of the outcome (`'p'`)

    >>> import pandas as pd
    >>> import matplotlib.pyplot as plt
    >>> from zepid.graphics import roc
    >>> df = pd.DataFrame()
    >>> df['d'] = [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1]
    >>> df['p'] = [0.1, 0.15, 0.1, 0.7, 0.5, 0.9, 0.95, 0.5, 0.4, 0.8, 0.99, 0.99, 0.89, 0.95]

    Creating ROC curve

    >>> roc(df, true='d', threshold='p', youden_index=False)
    >>> plt.show()
    """
    sens = []
    fpr = []
    thresh = []
    tf = df[[threshold, true]].copy()
    if tf.isnull().values.sum() != 0:
        raise ValueError('ROC curve cannot handle missing data for probability or true values')

    # Getting all possible cutpoints
    values = (list(np.unique(tf[threshold])))
    values = [float(np.min(tf[threshold]) - 0.001)] + values + [float(np.max(tf[threshold]) + 0.001)]
    # Going through all the cutpoints and calculating Sensitivity and 1-Specificity
    for v in list(reversed(values)):
        thresh.append(v)
        prediction = np.where(tf[threshold] >= v, 1, 0)
        se = prediction[tf[true] == 1].mean()
        sens.append(se)
        sp = prediction[tf[true] == 0].mean()
        fpr.append(sp)

    # If requested, calculate Youden's Index
    if youden_index is True:
        spec = [1 - i for i in fpr]
        youdens = []
        for i, j in zip(sens, spec):
            youdens.append(i + j - 1)
        ind = np.argmax(youdens)
        print('----------------------------------------------------------------------')
        print("Youden's Index: ", thresh[ind])
        print("Predictive values at Youden's Index")
        print("\tSensitivity: ", sens[ind])
        print("\tSpecificity: ", spec[ind])
        print('----------------------------------------------------------------------')

    # Creating ROC plot
    ax = plt.gca()
    ax.plot(fpr, sens, color='blue')
    ax.plot([0, 1], [0, 1], color='gray', linestyle='--')
    if youden_index is True:
        ax.text(0.65, 0.35, "Youden's Index:\n      " + str(round(thresh[ind], 5)))
    ax.set_xlim([-0.01, 1.01])
    ax.set_ylim([-0.01, 1.01])
    ax.set_ylabel('Sensitivity')
    ax.set_xlabel('1 -Specificity')
    return ax


def dynamic_risk_plot(risk_exposed, risk_unexposed, measure='RD', loess=True, loess_value=0.25, point_color='darkblue',
                      line_color='b', scale='linear'):
    """Creates a plot of how the risk difference or risk ratio changes over time with survival data. See the
    references for an example of this plot. Input data should be pandas Series indexed by 'timeline' where 'timeline'
    is the time corresponding to the risk estimate

    Parameters
    --------------
    risk_exposed : Series
        Pandas Series with the probability of the outcome among the exposed group. Index by 'timeline' where 'timeline'
        is the time. If you directly output the ``1 - survival_function_`` from lifelines.KaplanMeierFitter(), this
        should create a valid input
    risk_unexposed : Series
        Pandas Series with the probability of the outcome among the exposed group. Index by 'timeline' where 'timeline'
        is the time
    measure : str, optional
        Whether to generate the risk difference (RD) or risk ratio (RR). Default is 'RD'
    loess : bool, optional
        Whether to generate LOESS curve fit to the calculated points. Default is True
    loess_value : float, optional
        Fraction of values to fit LOESS curve to. Default is 0.25
    point_color : str, optional
        Color of the points
    line_color : str, optional
        Color of the LOESS line generated and plotted
    scale : str, optional
        Change the y-axis scale. Options are 'linear' (default), 'log', 'log-transform'. 'log' and 'log-transform' is
        only a valid option for Risk Ratio plots

    Returns
    -----------
    matplotlib axes

    Examples
    --------
    See graphics documentation or causal documentation for a detailed example.

    >>> import matplotlib.pyplot as plt
    >>> from zepid.graphics import dynamic_risk_plot

    >>> dynamic_risk_plot(a, b, loess=True)
    >>> plt.show()

    References
    ----------
    Cole SR, et al. (2014). Estimation of the standardized risk difference and ratio in a competing risks framework:
    application to injection drug use and progression to AIDS after initiation of antiretroviral therapy.
    AJE, 181(4), 238-245.
    """
    re = risk_exposed.drop_duplicates(keep='first').iloc[:, 0].rename('exposed').reset_index()
    ru = risk_unexposed.drop_duplicates(keep='first').iloc[:, 0].rename('unexposed').reset_index()
    re.timeline = np.round(re.timeline * 100000).astype(int) # This avoids a merge issue on floats
    ru.timeline = np.round(ru.timeline * 100000).astype(int)
    r = pd.merge(re, ru, how='outer', left_on='timeline', right_on='timeline').sort_values(by='timeline')
    r.timeline /= 100000
    r.ffill(inplace=True)
    if measure == 'RD':
        r['m'] = r['exposed'] - r['unexposed']
    elif measure == 'RR':
        r['m'] = r['exposed'] / r['unexposed']
        if scale == 'log-transform':
            r['m'] = np.log(r['m'])
    else:
        raise ValueError('Only "RD" and "RR" are currently supported')

    # Generating the plot
    ax = plt.gca()
    ax.plot(r['timeline'], r['m'], 'o', c=point_color)
    if loess is True:
        l = lowess(list(r['m']), list(r['timeline']), frac=loess_value)
        lowess_x = list(zip(*l))[0]
        lowess_y = list(zip(*l))[1]
        ax.plot(lowess_x, lowess_y, '-', c=line_color, linewidth=4)
    if measure == 'RD':
        ax.hlines(0, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
        ax.set_ylabel('Risk Difference')
    if measure == 'RR':
        if scale == 'log-transform':
            ax.hlines(0, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
            ax.set_ylabel('ln(Risk Ratio)')
        elif scale == 'log':
            ax.set_ylabel('Risk Ratio')
            ax.set_yscale('log')
            ax.yaxis.set_major_formatter(mticker.ScalarFormatter())
            ax.yaxis.get_major_formatter().set_scientific(False)
            ax.yaxis.get_major_formatter().set_useOffset(False)
            ax.hlines(1, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
        else:
            ax.hlines(1, 0, np.max(r['timeline'] + 0.5), linewidth=1.5)
    ax.set_xlabel('Time')
    ax.set_xlim([0, np.max(r['timeline']) + 0.5])
    return ax


def labbe_plot(r1=None, r0=None, scale='both', additive_tuner=12, multiplicative_tuner=12, figsize=(7, 4), **plot_kwargs):
    """L'Abbe plots are useful for summarizing measure modification on the difference or ratio scale. Primarily
    invented for meta-analysis usage, these plots display risk differences (or ratios) by their individual risks
    by an exposure. I find them most useful for a visualization of why if there is an association and there is no
    modfication on one scale (additive or multiplicative), then there must be modification on the other scale.

    Parameters
    ----------
    r1 : float, list, optional
        Single probability or a list of probabilities when exposure is 1. Default is None, which does not display points
    r0 : float, list, optional
        Single probability or a list of probabilities when exposure is 0. Default is None, which does not display points
    scale : str, optional
        Which scale to plot. The default is 'both', which generates side-by-side plots of additive scale and
        multiplicative scale. Other options are; 'additive' to display the additive plot, and 'multiplicative' to
        display the multiplicative plot
    additive_tuner : int, optional
        Optional parameter to change the number of lines displayed in the additive L'Abbe plot. Higher integer
        produces more reference lines
    multiplicative_tuner : int, optional
        Optional parameter to change the number of lines displayed in the multiplicative L'Abbe plot. Higher integer
        produces more reference lines
    figsize : set, optional
        Optional parameter to change the L'Abbe plot size. Only changes the plot size when scale='both'
    **plot_kwargs : optional
        Optional keyword arguments for matplotlib. kwargs will pass matplotlib.pyploy.plot kwargs are accepted. See
        matplotlib 'plot()' function documentation for further details

    Returns
    -------
    matplotlib axes

    Examples
    --------
    Setting up environment

    >>> import matplotlib.pyplot as plt
    >>> from zepid.graphics import labbe_plot

    Creating a blank plot

    >>> labbe_plot()
    >>> plt.show()

    Adding customized points to the plot

    >>> labbe_plot(r1=[0.3, 0.5], r0=[0.2, 0.7], scale='additive', color='r', marker='D', markersize=10, linestyle='')
    >>> plt.show()

    Only returning the additive plot

    >>> labbe_plot(r1=[0.3, 0.5], r0=[0.2, 0.7], scale='additive', markersize=10)
    >>> plt.show()

    Only returning the multiplicative plot

    >>> labbe_plot(r1=[0.3, 0.5], r0=[0.2, 0.7], scale='multiplicative', markersize=10)
    >>> plt.show()
    """
    if r1 is not None or r0 is not None:
        if len(list(r1)) != len(list(r0)):
            raise ValueError('The length of `r1` must be the same as `r0`')

    ya0 = np.linspace(0.0001, 0.9999, 12)

    if scale == 'both':
        fig, ax = plt.subplots(1, 2, figsize=figsize)
        ax[0].plot([0, 1], [0, 1], '--', color='gray', linewidth=1)
        for i in np.linspace(-1, 1, additive_tuner):
            ax[0].plot(ya0, ya0 + i, color='gray', linewidth=1)
        if r1 is not None and r0 is not None:
            if 'marker' in plot_kwargs:
                ax[0].plot(r0, r1, **plot_kwargs)
            else:  # When markers are unspecified, point estimates aren't displayed. This avoids
                ax[0].plot(r0, r1, 'o', **plot_kwargs)
        ax[0].set_xlim([0, 1])
        ax[0].set_ylim([0, 1])
        ax[0].set_yticks([0, 1])
        ax[0].set_xticks([0, 1])
        ax[0].set_xlabel("$P(Y|A=0)$")
        ax[0].set_ylabel("$P(Y|A=1)$")
        ax[0].set_title("Additive")

        ax[1].plot([0, 1], [0, 1], '--', color='gray', linewidth=1)
        for i in np.linspace(-3.5, 3.5, multiplicative_tuner):
            ax[1].plot(ya0, np.exp(np.log(ya0) + i), color='gray', linewidth=1)
        if r1 is not None and r0 is not None:
            if 'marker' in plot_kwargs:
                ax[1].plot(r0, r1, **plot_kwargs)
            else:  # When markers are unspecified, point estimates aren't displayed. This avoids
                ax[1].plot(r0, r1, 'o', **plot_kwargs)
        ax[1].set_xlim([0, 1])
        ax[1].set_ylim([0, 1])
        ax[1].set_yticks([])
        ax[1].set_xticks([0, 1])
        ax[1].set_xlabel("$P(Y|A=0)$")
        ax[1].set_title("Multiplicative")

    elif scale == 'additive':
        ax = plt.gca()
        ax.plot([0, 1], [0, 1], '--', color='gray', linewidth=1)
        for i in np.linspace(-1, 1, additive_tuner):
            ax.plot(ya0, ya0 + i, color='gray', linewidth=1)
        if r1 is not None and r0 is not None:
            if 'marker' in plot_kwargs:
                ax.plot(r0, r1, **plot_kwargs)
            else:  # When markers are unspecified, point estimates aren't displayed. This avoids
                ax.plot(r0, r1, 'o', **plot_kwargs)
        ax.set_xlim([0, 1])
        ax.set_ylim([0, 1])
        ax.set_yticks([0, 1])
        ax.set_xticks([0, 1])
        ax.set_xlabel("$P(Y|A=0)$")
        ax.set_ylabel("$P(Y|A=1)$")
        ax.set_title("Additive")

    elif scale == 'multiplicative':
        ax = plt.gca()
        ax.plot([0, 1], [0, 1], '--', color='gray', linewidth=1)
        for i in np.linspace(-3.2, 3.2, multiplicative_tuner):
            plt.plot(ya0, np.exp(np.log(ya0) + i), color='gray', linewidth=1)
        if r1 is not None and r0 is not None:
            if 'marker' in plot_kwargs:
                ax.plot(r0, r1, **plot_kwargs)
            else:  # When markers are unspecified, point estimates aren't displayed. This avoids
                ax.plot(r0, r1, 'o', **plot_kwargs)
        ax.set_xlim([0, 1])
        ax.set_ylim([0, 1])
        ax.set_yticks([0, 1])
        ax.set_xticks([0, 1])
        ax.set_xlabel("$P(Y|A=0)$")
        ax.set_ylabel("$P(Y|A=1)$")
        ax.set_title("Multiplicative")

    else:
        raise ValueError("`scale` must be either 'additive', 'multplicative', or 'both'")

    return ax


def zipper_plot(truth, lcl, ucl, colors=('blue', 'red')):
    """Zipper plots are a way to present simulation data, particularly confidence intervals and their width. They are
    also useful for showing the confidence interval coverage of the true parameter.

    Parameters
    ----------
    truth : float
        The true value with which to compare the confidence interval coverage to
    lcl : list, array, Series, container
        Container of lower confidence limits
    ucl : list, array, Series, container
        Container of upper confidence limits
    colors : set, list, container
        List of colors for confidence intervals. The first color is used to designate confidence intervals that cover
        the true value, and the second indicates confidence intervals

    Returns
    -------
    matplotlib axes

    Examples
    --------
    Setting up environment

    >>> import matplotlib.pyplot as plt
    >>> from zepid.graphics import zipper_plot

    Adding customized points to the plot

    >>> labbe_plot(r1=[0.3, 0.5], r0=[0.2, 0.7], scale='additive', color='r', marker='D', markersize=10, linestyle='')
    >>> plt.show()

    """
    # Formatting the input data
    dat = pd.DataFrame()
    dat['_lower_'] = lcl
    dat['_upper_'] = ucl
    if not np.all(dat['_lower_'] < dat['_upper_']):
        raise ValueError("It looks like some LCL are bigger than the UCL")
    dat['_cover_'] = np.where((dat['_lower_'] < truth) & (truth < dat['_upper_']), colors[0], colors[1])
    coverage = np.where((dat['_lower_'] < truth) & (truth < dat['_upper_']), 1, 0)
    dat = dat.sort_values(by=['_upper_', '_lower_']).reset_index(drop=True)

    # Formatting so it looks like a "zipper"
    spdat = np.split(dat, [int(dat.shape[0] / 2)], axis=0)
    spdat[0] = spdat[0].sort_values(by=['_upper_', '_lower_'], ascending=False)
    spdat[0]['_order_'] = list(range(0, dat.shape[0]-1, 2))
    spdat[1]['_order_'] = list(range(1, dat.shape[0]+1, 2))
    dat = pd.concat([spdat[0], spdat[1]]).sort_values(by='_order_')

    # Creating plot
    ax = plt.gca()
    ax.hlines(dat['_order_'], dat['_lower_'], dat['_upper_'], colors=dat['_cover_'])
    ax.vlines(truth, 0, dat.shape[0], colors='k')
    ax.set_ylim([0, dat.shape[0]])
    ax.set_yticks([])
    ax.set_xlabel("Confidence Intervals")
    ax.set_title("Estimated coverage: "+str(np.round(100*np.mean(coverage), 1))+"%")
    return ax



import warnings
import numpy as np
import pandas as pd

from zepid.causal.utils import propensity_score


class IPCW:
    r"""Calculates inverse probability of censoring weights. Note that this function will accept either a flat
    file (one row per individual) or a long format (multiple rows per individual). If a flat file is provided, it
    must be converted to a long format. This will be done automatically if `flat_df=True`. Additionally, a
    warning and some comparison statistics are provided. Please verify that they match. In general, it is recommended
    to convert the data set yourself

    IPCW are calculated via logistic regression and weights are cumulative products per unique ID. IPCW can
    be used to correct for missing at random data by the generated model in weighted Kaplan-Meier curves. The
    formula used to generate the unstabilized IPCW is

    .. math::

        \pi_i(t) = \prod_{R_k \le t} \frac{1}{\Pr(C_i > R_k | \bar{L} = \bar{l}, C_i > R_{k-1})}

    The stabilized IPCW substitutes predicted probabilities under the specified numerator model into the numerator
    of the previous equation. In general, it is recommended to stabilize IPCW by the time.

    .. math::

        \pi_i(t) = \prod_{R_k \le t} \frac{\Pr(C_i > R_k)}{\Pr(C_i > R_k | \bar{L} = \bar{l}, C_i > R_{k-1})}

    Note
    ----
    IPCW no longer support late-entry. The reason is that the pooled logistic regression model approach does not
    correctly accumulate the weights. As such, either all occurrences of late-entries need to be dropped (called the
    new-user design) or rows need to be back-propagated (unobserved rows are filled in). The second approach requires
    filling in the missing observed covariates and for time-varying variables will require imputation. The new-user
    design is a safer bet and generally what I will currently recommend

    Parameters
    ---------------
    df : DataFrame
        Pandas DataFrame object containing all the variables of interest
    idvar : str
        String that indicates the column name for a unique identifier for each individual
    time : str
        Column name for the ending observation time
    event : str
        Column name for the event of interest
    flat_df : bool, optional
        Whether the input dataframe only contains a single row per participant. If so, the flat dataframe is
        converted to a long dataframe. Default is False (for multiple rows per person)
    enter : str, optional
        Time participant began being observed. Default is None. This option is only needed when flat_df=True.
        Late-entries are no longer supported and specifying this will lead to a ValueError

    Example
    ------------
    Setting up the environment

    >>> from zepid import load_sample_data
    >>> from zepid.causal.ipw import IPCW
    >>> df = load_sample_data(timevary=True)
    >>> df['enter_q'] = df['enter'] ** 2
    >>> df['enter_c'] = df['enter'] ** 3
    >>> df['age0_q'] = df['age0'] ** 2
    >>> df['age0_c'] = df['age0'] ** 3

    Calculating stabilized IPCW with a long data set

    >>> ipc = IPCW(df, idvar='id', time='enter', event='dead')
    >>> ipc.regression_models(model_denominator='enter + enter_q + enter_c + male + age0 + age0_q + age0_c',
    >>>                       model_numerator='enter + enter_q + enter_c')
    >>> ipc.fit()

    Extracting calculated stabilized IPCW

    >>> ipc.Weight

    Calculating stabilized IPCW with a wide data set

    >>> df = load_sample_data(False)
    >>> ipc = IPCW(df, idvar='id', time='t', event='dead', flat_df=True)
    >>> ipc.regression_models(model_denominator='enter + enter_q + enter_c + male + age0 + age0_q + age0_c',
    >>>                       model_numerator='enter + enter_q + enter_c')
    >>> ipc.fit()

    References
    ----------
    Howe CJ et al. (2016) Selection bias due to loss to follow up in cohort studies. Epidemiology, 27(1), 91-97.
    """
    def __init__(self, df, idvar, time, event, flat_df=False, enter=None):
        if np.sum(df[time].isnull()) > 0:
            raise ValueError('Time is missing for at least one individual in the dataset. They must be removed before '
                             'IPCW can be fit.')
        f = df.copy()
        f.sort_values(by=[idvar, time], inplace=True)
        if np.max(f[time]) == 1:
            raise ValueError('The maximum observation time is 1. For IPCW to function properly, it needs to be greater'
                             'than 1. The more categories, the better the weight estimation')

        if flat_df:
            self.df = self._dataprep(f, idvar, time, event, enter=enter)
        else:
            late_check = f.drop_duplicates(subset=idvar, keep='first')
            if not np.all(late_check[time] <= 1):
                raise ValueError("IPCW no longer supports late-entries. IPCW cannot be correctly estimated using "
                                 "pooled logistic regression with late-entries. A future update will allow for weights "
                                 "with both late-entry and censoring")

            self.df = f
            self.df['__uncensored__'] = np.where((self.df[idvar] != self.df[idvar].shift(-1)) &
                                                 (self.df[event] == 0),
                                                 0, 1)  # generating indicator for uncensored
            self.df['__uncensored__'] = np.where(self.df[time] == np.max(df[time]), 1, self.df['__uncensored__'])

        self.idvar = idvar
        self.time = time
        self.event = event
        self.Weight = None

    def regression_models(self, model_denominator, model_numerator, print_results=True):
        """Regression model to generate predicted probabilities of censoring, conditional on specified variables.
        Whether stabilized or unstabilized IPCW are generated depends on the specified model numerator.

        Parameters
        --------------
        model_denominator : str
            String of predictor variables for the denominator following `patsy` syntax. Any variables included in the
            numerator, should be included in the denominator as well. Example 'var1 + var2 + var3 + t_start + t_squared'
        model_numerator : str
            String of predictor variables for the numerator following `patsy` syntax. In general, time is used as the
            stabilization factor. Example of argument 't_start + t_squared'
        print_results : bool, optional
            Whether to print the model results. Default is True
        """
        nmodel = propensity_score(self.df, '__uncensored__ ~ ' + model_numerator, print_results=print_results)
        self.df['__numer__'] = nmodel.predict(self.df)
        dmodel = propensity_score(self.df, '__uncensored__ ~ ' + model_denominator, print_results=print_results)
        self.df['__denom__'] = dmodel.predict(self.df)
        self.df['__cnumer__'] = self.df.groupby(self.idvar)['__numer__'].cumprod()
        self.df['__cdenom__'] = self.df.groupby(self.idvar)['__denom__'].cumprod()

    def fit(self):
        """Calculates IPCW for each observation period for each observation. The calculated weights can be
        accessed through the `IPCW.Weights` attribute

        Returns
        -------------
        Fills in the `Weight` attribute
        """
        self.Weight = self.df['__cnumer__'] / self.df['__cdenom__']

    @staticmethod
    def _dataprep(cf, idvar, time, event, enter=None):
        """Function to prepare the data to an appropriate format for the `IPCW` class. It breaks the dataset into
        single observations for event one unit increase in time. This is a background process and not meant for users
        to access

        Parameters for my reference
        cf:
            -pandas dataframe to convert into a long format
        idvar:
            -ID variable to retain for observations
        time:
            -last follow-up visit for participant
        event:
            -indicator of whether participant had the event (1 is yes, 0 is no)
        enter:
            -entry time for the participant. Default is None, which means all participants are assumed
             to enter at time zero. Input should be column name of entrance time
        """
        if enter is not None:
            raise ValueError("IPCW no longer supports late-entries. IPCW cannot be correctly estimated using pooled "
                             "logistic regression with late-entries. A future update will allow for weights with "
                             "both late-entry and censoring")

        # Copying observations over times
        cf['t_int_zepid'] = cf[time].astype(int)
        lf = pd.DataFrame(np.repeat(cf.values, cf['t_int_zepid'] + 1, axis=0), columns=cf.columns)
        lf['tpoint_zepid'] = lf.groupby(idvar)['t_int_zepid'].cumcount()
        lf['tdiff_zepid'] = lf[time] - lf['tpoint_zepid']
        lf = lf.loc[
            lf['tdiff_zepid'] != 0].copy()  # gets rid of censored at absolute time point (ex. censored at time 10)
        lf.loc[lf['tdiff_zepid'] > 1, 'delta_indicator_zepid'] = 0
        lf.loc[((lf['tdiff_zepid'] <= 1) & (lf[event] == 0)), 'delta_indicator_zepid'] = 0
        lf.loc[((lf['tdiff_zepid'] <= 1) & (lf[event] == 1) & (lf[idvar] != lf[idvar].shift(-1))),
               'delta_indicator_zepid'] = 1
        lf['t_enter_zepid'] = lf['tpoint_zepid']
        lf['t_out_zepid'] = np.where(lf['tdiff_zepid'] < 1, lf[time], lf['t_enter_zepid'] + 1)
        lf['uncensored_zepid'] = np.where((lf[idvar] != lf[idvar].shift(-1)) & (lf['delta_indicator_zepid'] == 0), 0, 1)
        lf['uncensored_zepid'] = np.where(lf['t_out_zepid'] == np.max(lf['t_out_zepid']), 1, lf['uncensored_zepid'])

        # Cleaning up the edited dataframe to return to user
        lf.drop(['tdiff_zepid', 'tpoint_zepid', 't_int_zepid', time, event], axis=1, inplace=True)
        lf.rename(columns={"delta_indicator_zepid": event, 'uncensored_zepid': '__uncensored__',
                           't_enter_zepid': 't_enter', 't_out_zepid': 't_out'}, inplace=True)

        warnings.warn('Please verify the long dataframe was generated correctly', UserWarning)
        print('Check for dataframe')
        print('\tEvents in input:', np.sum(cf[event]))
        print('\tEvents in output:', np.sum(lf[event]))
        print('\tCensor in input:', cf.dropna(subset=[event]).shape[0] - np.sum(cf[event]))
        print('\tCensor in output:', lf.shape[0] - np.sum(lf['__uncensored__']))
        print('\tTotal t input:', np.sum(cf[time]))
        print('\tTotal t output:', np.sum(lf.loc[(lf[idvar] != lf[idvar].shift(-1))]['t_out']))
        return lf.reset_index(drop=True)

